from __future__ import annotations

import ast
from os.path import exists, isdir

from simudator.core.processor import Processor

HELP_TEXT = """Here is a list of possible commands:
CLI control:
   - h, help : Shows this message.

   - q, quit : Closes the program. Nothing will be saved.

Simulation control:
   - n [x], next [x]: Simulates x ticks.
     If no x is given, simulates one tick.

   - rc, run_continuously:
     Continuously simulate ticks until a HALT is signalled
     or a breakpoint is reached.

   - u [x], undo [x]: Undoes the last x clock cycle.
     If no x is given one cycle is undone.

   - r, reset : Resets processor to initial state.

State information:
   - c, clock: Prints the processors current clock cycle.

   - p, print: Prints all states of all modules.

   - pl, printlist [add/remove] [module name] ...:
     Add or remove the given modules to a list of modules to print. If no
     arguments are given, i.e. "pl" or "printlist" on their own, all the modules 
     in the list will be printed.

   - pp, pretty print: Prints relevant processor state information in a
     readable and compact format.

   - ppv, pretty print verbose: Prints all processor state information in a
     readable and compact format.

Breakpoints:
   - p br, print breaks: Prints all breakpoints.

   - p br lam, print break lambdas:
     Prints all available lambdas of this processor to use for lambda
     breakpoints.

   - br state, break state [module name] [state name] [value]:
     Add a breakpoint for the module with the given name. The breakpoint brakes
     when the specified state of the module has the specified value.

   - br mem, break memory [module name] [address] [value]:
     Add a breakpoint for a memory module with the given name. The breakpoint
     breaks when the specified address of the memory has the specified value.

   - br lam, break lambda [lambda_name] [keyword argument] ...:
     Add a breakpoint that uses the referenced lambda to evaluate when to
     break. Any module names in the keyword arguments will be substituted with
     the actual module. Strings must be enclosed with quotation marks.

   - rm br [id], remove break [id]:
     Remove the breakpoint with the given ID.

File I/O:
   - l, load [path]:
     Loads a processors state from a given file path.

   - sf, save to file [path]:
     Saves the processor state to the specified file.
"""


class CLI:
    """
    Terminal based program for controlling a processor.

    Mostly used for testing and debugging.

    Parameters
    ----------
    processor : Processor
    """

    def __init__(self, processor: Processor):
        self._running = True
        self._processor = processor
        self._print_list = []

    def run(self) -> None:
        """Run the CLI.

        Put the user in an input loop where they can send commands to the
        program.
        """

        print("Welcome to simuDAtor!")

        while self._running:
            # Get user commands here
            user_input = input("> ")

            # Note: Keep the order of the cases consistent with the help text
            # for convenience
            match user_input.split():
                # CLI control -------------------------------------------------
                case ["h"] | ["help"]:
                    # prints help message
                    print(HELP_TEXT)

                case ["q"] | ["quit"]:
                    # Ends the loop
                    print("Goodbye")
                    self._running = False

                # Simulation control ------------------------------------------
                case ["n"] | ["next"]:
                    # runs next tick in the processor
                    # this could take time for larger simulations,
                    # could be good to add loading icon and stop user input

                    self._processor.do_tick()

                case ["n", _] | ["next", _]:
                    # runs a given number of ticks
                    # this could take time for larger simulations,
                    # could be good to add loading icon and stop user input

                    try:
                        for _ in range(int(user_input.split()[1])):
                            self._processor.do_tick()
                            if self._processor.is_stopped:
                                break
                        if self._processor.breakpoint_reached:
                            bp = self._processor.last_breakpoint
                            print(f"Reached breakpoint: {bp}")
                        if self._processor.should_halt():
                            print("The processor halted")
                    except ValueError:
                        print("Invalid value")

                case ["rc"] | ["run_continuously"]:
                    self._processor.run_continuously()
                    if self._processor.breakpoint_reached:
                        bp = self._processor.last_breakpoint
                        print(f"Reached breakpoint: {bp}")

                    if self._processor.should_halt():
                        print("The processor halted")

                case ["u"] | ["undo"]:
                    # undo one clock cycles

                    current_cycle = self._processor.get_clock()
                    try:
                        self._processor.load_cycle(max(current_cycle - 1, 0))
                    except IndexError:
                        print("Index out of range")

                case ["u", _] | ["undo", _]:
                    # undo many clock cycles
                    number = 1

                    try:
                        number = int(user_input.split()[1])
                    except ValueError:
                        print("Invalid value")
                        continue

                    current_cycle = self._processor.get_clock()
                    try:
                        self._processor.load_cycle(max(current_cycle - number, 0))
                    except IndexError:
                        print("Index out of range")

                case ["r"] | ["resets"]:
                    # self.processor.load_cycle(0)
                    self._processor.reset()

                # State information -------------------------------------------
                case ["c"] | ["clock"]:
                    # shows current clockcycle
                    print("Clock is at: ", str(self._processor.get_clock()))

                case ["p"] | ["print"]:
                    print("\n")
                    # prints all module states, is most prbably hard to read
                    for module in self._processor.get_modules():
                        module.print_module()
                        print("\n")

                case ["pl", *_] | ["printlist", *_]:
                    self._print_list_command(user_input.split())

                case ["pp"] | ["pretty print"]:
                    # pretty print the modules of the processor
                    self._processor.pretty_print()

                case ["ppv"] | ["pretty print verbose"]:
                    # pretty print the modules of the processor
                    # with all available information
                    self._processor.pretty_print_verbose()

                # Breakpoints -------------------------------------------------
                case ["p", "br"] | ["print", "breaks"]:
                    self._processor.print_breakpoints()

                case ["p", "br", "lam"] | ["print", "break", "lambdas"]:
                    lambdas = self._processor.get_breakpoint_lambdas()
                    if not lambdas:
                        print("There are no lambdas for this processor.")
                    else:
                        print(lambdas)

                case ["br", "state", _, _, _] | ["break", "state", _, _, _]:
                    module_name = user_input.split()[2]
                    state_name = user_input.split()[3]
                    val = ast.literal_eval(user_input.split()[4])
                    try:
                        self._processor.add_state_breakpoint(
                            module_name, state_name, val
                        )
                    except ValueError as e:
                        print(e)

                case ["br", "mem", _, _, _] | ["break", "memory", _, _, _]:
                    module_name = user_input.split()[2]
                    try:
                        address = ast.literal_eval(user_input.split()[3])
                        value = ast.literal_eval(user_input.split()[4])
                        self._processor.add_memory_breakpoint(
                            module_name, address, value
                        )
                    except ValueError as e:
                        print(e)

                case ["br", "lam", _, *_] | ["break", "lambda", _, *_]:
                    lambda_name = user_input.split()[2]
                    kwargs_str = user_input.split()[3:]
                    kwargs = {}
                    for kwarg in kwargs_str:
                        key, value = kwarg.split('=')
                        kwargs[key] = ast.literal_eval(value)
                    try:
                        self._processor.add_lambda_breakpoint(lambda_name, **kwargs)
                    except ValueError as e:
                        print(e)

                case ["rm", "br", _] | ["remove", "break", _]:
                    try:
                        bp_id = int(user_input.split()[2])
                        if not self._processor.remove_breakpoint(bp_id):
                            print(f"No breakpoint with ID {bp_id}")
                    except ValueError:
                        print("Breakpoint ID must be an integer")

                # File I/O ----------------------------------------------------
                case ["l", _] | ["load", _]:
                    # load processor state from file
                    try:
                        self._processor.reset()
                        self._processor.load_state_from_file(user_input.split()[1])
                    except FileNotFoundError:
                        print("No file called ", user_input.split()[1], " was found")

                case ["sf"] | ["save to file"]:
                    file_path = "cpu_save_file"
                    self._processor.save_state_to_file(file_path)

                case ["sf", _] | ["save to file", _]:
                    file_path = user_input.split()[1]
                    self._save_to_file(file_path)

                # -------------------------------------------------------------
                case _:
                    # message for unknown command
                    print("Unknown command. Write \"help\" for assistance.")

    def _save_to_file(self, file_path: str) -> None:
        """Save the processor state to the given file path.

        No save is made if the file path is a directory. If the file already
        exists, the user is given the choice to overwrite the file.

        Parameters
        ----------
        file_path : str
        """

        # Given file does not exist, save_state_to_file will create it
        if not exists(file_path):
            self._processor.save_state_to_file(file_path)

        else:
            # We dont touch directories
            if isdir(file_path):
                print("Given file is a directory")
                return

            # Should we overwrite file?
            print("Given file already exists, do you want to overwrite it?")
            overwrite = ""
            while True:
                overwrite = input("yes/no: ")
                if overwrite in ["yes", "no"]:
                    break
                else:
                    print("Please give an input of 'yes' or 'no'!")

            if overwrite == "yes":
                self._processor.save_state_to_file(file_path)

    def _print_list_command(self, user_input: list[str]) -> None:
        """Add/remove modules to internal list or print all modules in the
        internal list.

        Parameters
        ----------
        user_input : list[str]
            A print list command split by spaces. Should be on the format
            ``["printlist", "add", MODULE_NAME, ...]`` or
            ``["printlist", "remove", MODULE_NAME, ...]`` to add or remove
            modules, where MODULE_NAME is the name of a module as a string,
            or ``["printlist"]`` to print all added modules.
            (``"printlist"`` can be substituted with ``"pl"``)
        """
        # no additional commands so only print
        if len(user_input) == 1:
            print("\n")
            for name in self._print_list:
                self._processor.get_module(name).print_module()
                print("\n")
            return

        names = user_input[2:]
        # add name to print list
        if user_input[1] == "add":
            for name in names:
                # check if name exists
                try:
                    self._processor.get_module(name)
                    if name not in self._print_list:
                        self._print_list.append(name)
                except KeyError:
                    print(
                        "No module named",
                        name,
                    )

        elif user_input[1] == "remove":
            for name in names:
                if name in self._print_list:
                    self._print_list.remove(name)
                else:
                    print(name, "not in print list")
        else:
            print("Unknown command")


if __name__ == '__main__':
    cli = CLI(None)
    cli.run()
