from typing import Any

from simudator.core.module import Module
from simudator.core.signal import Signal


class Demux(Module):
    """
    A general demultiplexer that allows an arbitrary amount of outputs to map
    to a single input.

    Assumes the value sent from input to outputs is an integer.

    Parameters
    ----------
    control : Signal
        Control signal of which the value is used to index which
        output signal to output to.
    bit_length : int
        Maximum number of bits for the value outputted by the demux. All extra
        bits of the input value are discarded.
    input : Signal
        Input signal of which the value is forwarded to the currently
        selected output signal.
    outputs : list[Signal]
        List of output signals to demultiplex the input signal value to.
    name : str
        Name of the demultiplexer module.
    value : Any
        Initial value to output to the currently selected output signal.
    """

    __slots__ = ("_bit_length", "_value")

    def __init__(
        self,
        control: Signal,
        bit_length: int,
        input: Signal,
        outputs: list[Signal] = [],
        name="demux",
        value=0,
    ) -> None:

        # Signals
        signals = {
            "in_control": control,
            "in_input": input,
        }
        for i, s in enumerate(outputs):
            signals[f"out_output_{i}"] = s

        super().__init__(signals, name)

        self._bit_length = bit_length

        # Value to be read/written
        self._value = value

    def update_register(self) -> None:
        """
        Read the input signal value and update the value of the demux.
        """

        input_value = self.signals["in_input"].get_value()
        mask = 2**self._bit_length - 1
        self._value = input_value & mask

    def output_register(self) -> None:
        """Read which signal to write to from the control signal and forward
        the input to that output signal.
        """
        output_index = self.signals["in_control"].get_value()
        output_signal_key = f"out_output_{output_index}"
        self.signals[output_signal_key].update_value(self._value)

    def update_logic(self) -> None:
        """Do nothing.

        The demux has no logic.
        """
        pass

    def get_state(self) -> dict[str, Any]:
        state = super().get_state()
        state["value"] = self._value
        state["bit_length"] = self._bit_length

        return state

    def set_state(self, state: dict[str, Any]) -> None:
        """Set the name and value of the demux.

        Parameters
        ----------
        state : dict[str, Any]
            The state of the demux to load. Should contain the keys "value" and
            "bit_length", both with values of type ``int``.
        """
        super().set_state(state)
        self._value = state["value"]
        self._bit_length = state["bit_length"]
