from typing import Any

from simudator.core.module import Module
from simudator.core.signal import Signal


class Mux(Module):
    """
    A general mux that allows an arbitrary amount of input to map to a single
    output.

    Parameters
    ----------
    control : Signal
        Control signal of which the value is used to index which
        input signal to read from.
    bit_length : int
        Maximum number of bits for the value outputted by the mux. All extra
        bits of the input value are discarded.
    inputs : list[Signal]
        List of input signals to select from to output onto the output signal.
    output : Signal
        Signal onto which the value of the currently selected input is
        outputted.
    name : str
        Name of the multiplexer module.
    value : Any
        Initial value to output to the output signal.
    """

    __slots__ = ("_bit_length", "_value")

    def __init__(
        self,
        control: Signal,
        bit_length: int,
        output: Signal,
        inputs: list[Signal] = [],
        name="mux",
        value=0,
    ) -> None:

        # Signals
        signals = {
            "in_control": control,
            "out": output,
        }
        for i, s in enumerate(inputs):
            signals[f"in_input_{i}"] = s

        # Init super class
        super().__init__(signals, name)

        # mask and bit_length
        self._bit_length = bit_length

        # Value to be read/written
        self._value = value

    def update_register(self) -> None:
        """Read which signal to read from the control signal and forward that
        signal to the output.
        """
        input_index = self.signals["in_control"].get_value()
        input_signal_key = f"in_input_{input_index}"
        input_value = self.signals[input_signal_key].get_value()
        mask = 2**self._bit_length - 1
        self._value = input_value & mask

    def output_register(self):
        """Output the value of the currently selected input signal to the
        output signal.
        """
        self.signals["out"].update_value(self._value)

    def update_logic(self):
        """Do nothing.

        The mux has no logic.
        """
        pass

    def get_state(self) -> dict[str, Any]:
        state = super().get_state()
        state["value"] = self._value
        state["bit_length"] = self._bit_length

        return state

    def set_state(self, state: dict[str, Any]) -> None:
        """Set the name, value and bit length of the demux.

        Parameters
        ----------
        state : dict[str, Any]
            The state of the demux to load. Should contain the keys "name",
            "value" and "bit_length" with values of type ``str``, ``int`` and
            ``int`` respectively.
        """
        super().set_state(state)
        self._value = state["value"]
        self._bit_length = state["bit_length"]

    def print_module(self) -> None:
        print(
            self.name,
            "\n-----",
            "\nvalue: ",
            self._value,
            "\nbit length: ",
            self._bit_length,
        )
