import os
from collections.abc import Callable
from typing import Any

from simudator.core.breakpoint import Breakpoint
from simudator.core.breakpoint_lambda import LambdaBreakpoint
from simudator.core.breakpoint_memory import MemoryBreakpoint
from simudator.core.breakpoint_state import StateBreakpoint
from simudator.core.module import Module
from simudator.core.modules.memory import Memory
from simudator.processor.mia.modules.bus import Bus
from simudator.processor.mia.modules.micro_memory import MicroMemory

from .signal import Signal


class Processor:
    """
    Main class for controlling the processor.

    Uses modules and signals to simulate processor components behaviour.
    """

    __slots__ = (
        "modules",
        "signals",
        "clock",
        "update_queue",
        "module_history",
        "signal_history",
        "breakpoint_id_counter",
        "breakpoints",
        "breakpoint_reached",
        "last_breakpoint",
        "cycles_to_save",
        "removed_cycles",
        "max_line_len",
        "line_separator",
        "is_stopped",
        "new_instruction",
        "current_instructions",
        "ignore_keys",
        "lambdas",
        "assembly_cycles",
        "line_separator",
    )

    LINE_SEPARATOR = "-"
    MAX_LINE_LEN = 170
    CYCLES_TO_SAVE = 10000

    def __init__(self) -> None:
        # For simulation
        self.modules: dict[str, Module] = dict()
        self.signals = []
        self.clock = 0
        self.update_queue = []
        self.is_stopped = False

        # Breakpoint handling
        self.breakpoint_id_counter = 1
        self.breakpoints: dict[int, Breakpoint] = {}
        self.breakpoint_reached = False
        self.last_breakpoint = None
        self.lambdas: dict[str, Callable[..., bool]] = {}

        # Saving processor state for saving/loading to file and undoing ticks
        self.removed_cycles = 0
        self.assembly_cycles = [0]  # Map asm instruction to clock cycle
        self.module_history: list[dict[str, dict[str, Any]]] = []
        self.signal_history: list[list] = []  # TODO: Is this needed?

        # For showing which instructions are being done and signalling
        # the start of a new one
        self.new_instruction = False

        # It is the responsibility of the CPU to tell the gui:s pipeline
        # diagram what information to be displayed and how to display it.
        # Thus each CPU keeps track of its current instructions together with
        # each instructions position in the pipeline diagram.
        self.current_instructions: list[tuple[str, int, int]] = []

        # TODO: keeping track of what pieces of info not to show
        # show not be done at the processor level.
        # Maybe implement a 'get_pretty_print_state' at module
        # level?
        self.ignore_keys = [
            "bit_length",
            "mask",
            "increment",
            "read_from_bus",
            "read_from_uADR",
            "decrement_by_one",
            "bus_id",
        ]

    def do_tick(self) -> None:
        """Simulate one clock cycle of the processor.

        Run each module's tick function, then handle the following updates.
        Also check for breakpoints that are reached in this cycle, and save
        clock cycles when new assembly instructions are started.
        """
        if len(self.module_history) > self.clock - self.removed_cycles:
            # If a previous stored cycle has been loaded, discard
            # all stored cycles from that cycle and onward in
            # the history of saved cycles
            self.module_history = self.module_history[0 : self.clock]
            self.signal_history = self.signal_history[0 : self.clock]

        self.unstop()
        self.save_cycle()
        self.clock += 1

        for module in self.modules.values():
            module.update_register()

        for module in self.modules.values():
            module.output_register()

        while self.update_queue:
            module = self.update_queue.pop(0)
            module.update_logic()

        self.stop_at_breakpoints()
        if self.should_halt():
            self.stop()

        # Set new_instruction and save clock cycle for each
        # new assembly instruction
        if self.is_new_instruction():
            self.new_instruction = True
            self.current_instructions = self.get_current_instructions()
            self.assembly_cycles.append(self.get_clock())
        else:
            self.new_instruction = False

    def get_current_instructions(self) -> list[tuple[str, int, int]]:
        """Return a list of the current instructions with their positions.

        Each entry in the list is a tuple of a instruction with its column and
        row position for the pipeline diagram.

        Returns
        -------
        list[tuple[str, int, int]]
            A list of instructions represented as tuples. Each tuple contains
            the instruction and its position in the pipeline.
        """
        raise NotImplementedError

    def get_pipeline_dimensions(self) -> tuple[int, int]:
        """Return the dimensions of the pipeline of the processor.

        Returns
        -------
        width : int
            The width of the pipeline (number of cycles to display at a time).
        depth : int
           Depth of the pipeline.
        """
        raise NotImplementedError

    def get_asm_instruction(self) -> str:
        """Return a string containing the 'code' for the current assembly
        instruction.

        What 'code' refers to is up to each processor to define. It could be a
        memory address, an instruction label or something else.

        Returns
        -------
        str
            The current assembly instruction.
        """
        raise NotImplementedError

    def run_asm_instruction(self, num_instructions=1) -> None:
        """Run a number of assembly instructions.

        Defaults to running 1 instruction unless some other number is provided.

        Parameters
        ----------
        num_instructions : int
            Number of assembly instructions to perform.

        Notes
        -----
        This should be implemented per processor and thus we
        raise NotImplementedError.
        """

        raise NotImplementedError

    def undo_asm_instruction(self, num_instructions=1) -> None:
        """
        Undo a number of assembly instructions on the CPU.

        Default argument is one, but it is possible to specify any number of
        instructions.

        Parameters
        ----------
        num_instructions : int
            Number of assembly instructions to undo.
        """

        current_clock_cycle = self.clock
        index = len(self.assembly_cycles)
        saved_cycle = self.assembly_cycles[index - 1]

        # Make sure we are undoing the instruction(s) we are currently on
        while saved_cycle >= current_clock_cycle:
            index -= 1
            saved_cycle = self.assembly_cycles[index - 1]

        index -= num_instructions

        # We cant undo more instructions than we have performed.
        if index < 0:
            raise IndexError

        clockcycle = self.assembly_cycles[index]

        self.load_cycle(clockcycle)

        # Need +1 here since we save the start state to enable to
        # load the start state. This is done since we only append clock
        # cycles to the list self.assembly_cycles when we reach a new state
        # that has uPC set to 0, which wont happen when we load a new file.
        self.assembly_cycles = self.assembly_cycles[: index + 1]

    def run_continuously(self) -> None:
        """
        Run the processor until it halts, is stopped or reaches a breakpoint.
        """
        self.unstop()
        while not self.is_stopped:
            self.do_tick()

    def should_halt(self) -> bool:
        """
        Return whether the processor should halt this clock cycle or not.

        Returns
        -------
        bool
            ``True`` if the processor should halt, ``False`` otherwise.

        Notes
        ----
        Each processor should implement the condition for True in its subclass.
        If we instead raise NotImplementedError as we do in all other cases
        when the functionality should be implemented in the subclass we lose
        the ability to test modules on the base processor class.
        """
        return False

    def stop(self) -> None:
        """Signal to stop the execution of the processor until the processor is
        instructed to run continuously or do some ticks again.
        """
        self.is_stopped = True

    def unstop(self) -> None:
        """Reset the stop execution signal."""
        self.is_stopped = False

    def is_new_instruction(self) -> bool:
        """
        Return true on the same clock cycle the CPU starts a new instruction.

        Should be implemented per CPU, return field new_instruction. It is up
        to each processor to set this field to True/False correctly.
        """
        return self.new_instruction

    def stop_at_breakpoints(self) -> None:
        """Stop the execution if any breakpoint has been reached during this
        clock cycle.

        Also record the breakpoint that was reached.
        """
        self.breakpoint_reached = False
        for _, bp in self.breakpoints.items():
            # TODO: Can make this more efficient by only checking enabled
            # breakpoints
            if bp.is_break() and bp.is_enabled:
                self.breakpoint_reached = True
                self.last_breakpoint = bp

        if self.breakpoint_reached:
            self.stop()

    def reset(self):
        """Reset the processor to its start state.

        This resets all modules and removes any saved states for undoing clock
        cycles. A round of value propagation is done to reset signals too.
        """
        self.clock = 0
        self.module_history.clear()
        self.signal_history.clear()
        self.removed_cycles = 0
        self.assembly_cycles = [0]

        for module in self.modules.values():
            module.reset()
            module.output_register()

        while self.update_queue:
            module = self.update_queue.pop(0)
            module.update_logic()

    def add_modules_to_update(self, module: Module) -> None:
        """Queue a module to be updated at the end of clock cycle.

        Parameters
        ----------
        module : Module
            Module to queue for update.

        See Also
        --------
        Module.update_logic :
            Method for updating a module.
        """
        if module not in self.update_queue:
            self.update_queue.append(module)

    def add_module(self, module: Module) -> None:
        """Add module to be simulated by the processor.

        Parameters
        ----------
        module : Module
            Module to add.
        """
        self.modules[module.name] = module

    def get_module(self, name: str) -> Module:
        """Get module with specific name.

        Parameters
        ----------
        name : str
            Name of the module to get.

        Returns
        -------
        Module
            The module with the specified name.
        """
        return self.modules[name]

    def get_modules(self) -> list[Module]:
        """Get list of all modules.

        Returns
        -------
        list[Module]
            List of all modules in the processor.
        """
        return list(self.modules.values())

    def add_signals(self, signals: list[Signal]) -> None:
        """Add signals to the processor.

        Parameters
        ----------
        list[Signal]
            List of signals to add for simulation.
        """
        self.signals += signals

    def get_clock(self) -> int:
        """Get the current clockcycle number.

        Returns
        -------
        int
            Current clock cycle number of the processor.
        """
        return self.clock

    def set_clock(self, value: int) -> None:
        """Set current clockcycle number.

        Parameters
        ----------
        value : int
            Cycle number to set the clock to.
        """
        self.clock = value

    def save_cycle(self) -> None:
        """
        Save the current state for all the modules and signals in the
        processor to the processor's internal history of all clock cycles.
        """
        module_states = {}

        # Only save a specified number of cycles,
        # saving every cycle can easily eat all ram
        if len(self.module_history) > Processor.CYCLES_TO_SAVE:
            self.module_history.pop(0)
            self.removed_cycles += 1

        for module in self.modules.values():
            module_states[module.name] = module.get_state()
        self.module_history.append(module_states)

    def load_cycle(self, cycle: int) -> None:
        """Load the state of all modules as they were at the specified clock
        cycle.

        This does not erase the history of states for the later cycles.

        Parameters
        ----------
        cycle : int
            Number of the cycle to load.
        """

        cycle_index = cycle - self.removed_cycles

        if cycle_index < 0:
            raise ValueError("The cycle to be loaded is not saved")

        try:
            module_states = self.module_history[cycle_index]
        except IndexError:
            raise IndexError

        for module_name, module_state in module_states.items():
            self.modules[module_name].set_state(module_state)

        self.clock = cycle

        for module in self.modules.values():
            module.output_register()

        while self.update_queue:
            module = self.update_queue.pop(0)
            module.update_logic()

    def load_state_from_file(self, file_path: str) -> None:
        """Load states for modules from a file.

        Parameters
        ----------
        file_path : str
            Path to the file containing the processor state to load.
        """
        self.reset()

        file = open(file_path)

        # Ensure we are at the start of the file
        file.seek(0, os.SEEK_END)  # Go to end of file
        if file.tell():  # If current pos != 0
            file.seek(0)  # Return to the start of the file
        else:
            raise ValueError  # Else raise error

        # Appends the lines from the loaded file to a long string. This string
        # is given to each module in the function 'load_from_str'. Each module
        # will thus load itself.
        module_name = None
        module_str = ""

        # Go through each line in the file
        lines = file.readlines()
        for line in lines:

            # If no name currently saved then get name from current line
            if module_name is None:
                module_name = line.partition(":")[0]

            # if already has name get line content
            else:
                # if not empty line, then save content
                if line.strip():
                    module_str += line

                # if line is empty then give saved content to name file
                else:
                    module = self.get_module(module_name)
                    module.load_from_str(module_str)
                    module_str = ""
                    module_name = None

        # The last module will be an edge case, since we append the information
        # after we read an empty line.
        if module_name and module_str:
            module = self.get_module(module_name)
            module.load_from_str(module_str)

    def save_state_to_file(self, file_path: str) -> None:
        """Save the current state of the cpu to file.

        Will overwrite the given file.

        Parameters
        ----------
        file_path : str
            Path to the file to write to.
        """
        file = open(file_path, "w")
        file.write("")
        file.close()

        for module in self.modules.values():
            print(module)
            module.save_state_to_file(file_path)

    def pretty_print(self) -> None:
        """Print the processor state in a readable and compact format."""
        self.pretty_print_verbose(self.ignore_keys)

    def pretty_print_verbose(self, ignore_keys=None) -> None:
        """
        Print the most relevant information about each module in a compact and
        readable format.

        State variables of modules can be ignored with the optional argument.

        Parameters
        ----------
        ignore_keys : list[str]
            List of names of state variables of modules to exclude when
            printing module states.
        """
        # TODO: ignore keys per module and not for all modules
        memory_modules = []
        other_modules = []

        if ignore_keys is None:
            ignore_keys = []

        # TODO: remove isinstance(module, micro_memory)
        for module in self.modules.values():
            if not isinstance(module, Bus):
                if isinstance(module, Memory) or isinstance(module, MicroMemory):
                    memory_modules.append(module)
                else:
                    other_modules.append(module)

        # sort the modules by name to ensure that they appear in the
        # same order
        memory_modules.sort(key=lambda x: x.name, reverse=True)
        other_modules.sort(key=lambda x: x.name, reverse=True)

        # specify which keys to ignore from the modules 'get_state'
        # function
        # TODO: ignore fields per module, this will ignore 'mask'
        # for every module

        module_to_line_length = {}
        # get the longest line length for all other_modules
        for module in other_modules:
            # +3 for padding
            line_len = module.get_longest_line_len(ignore_keys) + 3

            # TODO: what to do if two or more modules has the same name
            module_to_line_length[module] = line_len

        groups = self.group_pp_modules(module_to_line_length)

        print(Processor.LINE_SEPARATOR * Processor.MAX_LINE_LEN)

        # print each group separate
        for group in groups:
            self.pretty_print_names(group)

            # Build a new string containing the information to show
            # the user
            for row in range(self.get_most_fields(group, ignore_keys)):
                string = ""
                total_padding = 0

                # Get the information from each module
                for module in group:
                    module_state = module.get_state()
                    keys = self.fields_to_list(module)

                    # Get the keys that we want to print to the user
                    # If this is not done, the keys we want to print
                    # could end up on an index that is higher than the
                    # number of rows printed and will therefore be missed
                    real_keys = [key for key in keys if key not in ignore_keys]

                    # Dont go out of index
                    # Needed since one module might want to print 4
                    # fields and another only 1
                    if row < len(real_keys) and real_keys[row] not in ignore_keys:
                        string += (
                            real_keys[row] + ": " + str(module_state[real_keys[row]])
                        )

                    # pad the string so each string has the same length
                    total_padding += module_to_line_length[module]
                    string = string.ljust(total_padding)
                    # replace last to chars with a separator and padding
                    string = string[0:-2] + "| "

                print(string)
            print(Processor.LINE_SEPARATOR * Processor.MAX_LINE_LEN)

        for memory_module in memory_modules:
            self.pretty_print_memory(memory_module)

    def pretty_print_memory(self, module: Memory) -> None:
        """Print a memory module in a compact and readable format.

        Parameters
        ----------
        module : Memory
            Memory module to print the state of.
        """
        print(module.name)
        print(Processor.LINE_SEPARATOR * Processor.MAX_LINE_LEN)

        longest_line_len = module.get_longest_line_len()
        # longest_line_len = self.get_longest_memory_value(module.memory)
        string = ""
        last_mem_len = module.get_largest_mem_adr()

        for i, value in enumerate(module.get_state()["memory"]):

            # create a new string containing the address and the value
            # of the memory address formatted to fit with the largest
            # address and largest memory value
            new_row = (str(i).ljust(last_mem_len) + ": " + str(value)).ljust(
                longest_line_len + last_mem_len + 3
            ) + "|"

            # only add the string if there is space for it, else
            # print the string and start a new
            if len(string + new_row) + 1 > Processor.MAX_LINE_LEN:
                print(string)
                string = new_row
            else:
                # First iteration string will be
                # empty and should not be padded
                if string:
                    string += " " + new_row
                else:
                    string = new_row
        print(string)
        print(Processor.LINE_SEPARATOR * Processor.MAX_LINE_LEN)
        print()

    def pretty_print_names(self, module_to_line_length: dict[Module, int]) -> None:
        """
        Print the name of the modules in one row with formatting.

        Adds spacing between the names so that the longest state variable of
        each module has space to be printed below the name.

        Parameters
        ----------
        module_to_line_length : dict[Module, int]
            Mapping from module to length of the longest state variable of
            the module.
        """
        name_string = ""
        total_len = 0
        for module in module_to_line_length:
            name_string += module.name
            total_len += module_to_line_length[module]
            name_string = name_string.ljust(total_len)
            name_string = name_string[0:-2] + "| "

        print(name_string)

    def group_pp_modules(
        self, module_to_line_length: dict[Module, int]
    ) -> list[dict[Module, int]]:
        """Group the modules to be pretty printed into groups with a
        total line length lower than 'Processor.MAX_LINE_LEN'.

        Parameters
        ----------
        module_to_line_length : dict[Module, int]
            Mapping from module to length of the longest state variable of
            the module (takes as the line length).

        Returns
        -------
        list[dict[Module, int]]
            List of mappings from module to line length, each mapping
            representing a group of modules for printing.
        """

        groups = [{}]
        group_index = 0
        line_len = 0
        for module in module_to_line_length:

            # Make sure the line is not to long
            if line_len + module_to_line_length[module] < Processor.MAX_LINE_LEN:
                line_len += module_to_line_length[module]
                groups[group_index][module] = module_to_line_length[module]

            # If it would have been, start a new group
            else:
                groups.append({})
                group_index += 1
                groups[group_index][module] = module_to_line_length[module]
                line_len = module_to_line_length[module]
        return groups

    def fields_to_list(self, module: Module, ignore_keys=[]) -> list[str]:
        """
        Return a list containing all state variable names, excluding module
        name, for a module.
        Optional argument to ignore specific state variables.

        Parameters
        ----------
        module : Module
            Module to get state variable names from.
        ignore_keys : list[str]
            List of state variable names to exclude.

        Returns
        -------
        list[str]
            List of state variable names, excluding module name, of a module.
        """
        return [
            key
            for key in module.get_state()
            if key != "name" and key not in ignore_keys
        ]

    def get_most_fields(self, modules: dict[Module, int], ignore_keys=None) -> int:
        """Get the maximum number of state variables among all modules.

        Can optionally ignore keys.

        Parameters
        ----------
        modules : dict[Module, int]
            Mapping from module to length of the longest state variable of
            the module. Used as a list of modules to take the maximum over.
        ignore_keys : list[str]
            State variables to exclude.

        Returns
        -------
        int
            The maximum of the number of module state variables among all
            modules of the processor.
        """

        if ignore_keys is None:
            ignore_keys = []

        fields = 0
        for module in modules:
            module_fields = len(self.fields_to_list(module, ignore_keys))
            if module_fields > fields:
                fields = module_fields

        return fields

    def add_state_breakpoint(
        self, module_name: str, state_name: str, value: Any
    ) -> None:
        """Add a module state breakpoint to the processor.

        Parameters
        ----------
            module_name : str
                Name of the module for which a breakpoint will be added.
            state_name : str
                Name of the state variable of the module to put a breakpoint
                on.
            value : Any
                Value of the state variable that triggers the breakpoint.

        Raises
        ------
        ValueError
            If the module does not exist or if the state variable does not
            exist in the module.
        """

        if module_name not in self.modules:
            raise ValueError(f"No module named {module_name}")

        module = self.modules[module_name]

        if state_name not in module.get_state():
            raise ValueError(f"No state named {state_name} in " f"module {module_name}")

        bp = StateBreakpoint(module, state_name, value)
        self.breakpoints[self.breakpoint_id_counter] = bp
        self.breakpoint_id_counter += 1

    def remove_breakpoint(self, id: int) -> bool:
        """Remove a specific breakpoint.

        Parameters
        ----------
        id : int
            ID of the breakpoint to be removed.
        """
        if id in self.breakpoints:
            del self.breakpoints[id]
            return True
        return False

    def print_breakpoints(self) -> None:
        """Print all breakpoints of the processor."""
        if not self.breakpoints:
            print("There are no breakpoints.")
        else:
            for bp_id, bp in self.breakpoints.items():
                print(f"BP {bp_id}: {bp}")

    def add_memory_breakpoint(self, module_name: str, address: int, value: Any) -> None:
        """Add a breakpoint for a memory module.

        Parameters
        ----------
        module_name : str
            Name of the memory module for which a breakpoint will be added.
        address : int
            Address of the memory module to break at.
        value : Any
            Value of the address that triggers the breakpoint.

        Raises
        ------
        ValueError
            If the module does not exist or is not a memory.
        """

        if module_name not in self.modules:
            raise ValueError(f"No module named {module_name}")

        module = self.modules[module_name]
        if "memory" not in module.get_state():
            raise ValueError(f"Module {module_name} is not a memory.")

        bp = MemoryBreakpoint(module, address, value)
        self.breakpoints[self.breakpoint_id_counter] = bp
        self.breakpoint_id_counter += 1

    def get_breakpoint_lambdas(self) -> list[str]:
        """Get all functions available for lambda breakpoints.

        Returns
        -------
        list[str]
            List of names of all functions of the processor that are available
            for adding lambda breakpoints.
        """

        return list(self.lambdas.keys())

    def add_lambda_breakpoint(self, lambda_name: str, **kwargs) -> None:
        """Add a lambda breakpoint to the processor.

        Module names used in the keyword arguments are substituted by the
        actual modules before being passed on to the lambda breakpoint
        function.

        Parameters
        ----------
        lambda_name : str
            Name of the lambda breakpoint function to be used by the
            breakpoint.
        **kwargs
            Arguments to pass to the breakpoint function.

        Raises
        ------
        ValueError
            If the  lambda breakpoint function does not exist.
        """
        # Substitute any module names for the module instances
        # This is to be able to externally create general lambda breakpoints
        # that can take any values as arguments while also having a way to take
        # modules as arguments without having direct access to them
        for key in kwargs.keys():
            value = kwargs[key]
            if value in self.modules:
                kwargs[key] = self.modules[value]

        if lambda_name not in self.lambdas:
            raise ValueError(f"No lambda named {lambda_name}.")

        lambda_func = self.lambdas[lambda_name]
        bp = LambdaBreakpoint(lambda_func, **kwargs)
        self.breakpoints[self.breakpoint_id_counter] = bp
        self.breakpoint_id_counter += 1

    def set_enabled_breakpoint(self, bp_id: int, is_enabled: bool) -> None:
        """Toggle a breakpoint to enabled or disabled.

        Parameters
        ----------
        bp_id : int
            ID of the breakpoint to toggle.
        is_enabled : bool
            ``True`` to enable, ``False`` to disable the breakpoint.
        """
        self.breakpoints[bp_id].set_enabled(is_enabled)
