from __future__ import annotations

from typing import Any


class Signal:
    """
    Signal class to send data between modules in a processor.

    Parameters
    ----------
    processor : Processor
        Processor that the signal belongs to. Needed to queue processor modules
        for update.
    name : str
        Name of the signal.
    value : Any
        Initial value of the signal.
    """

    __slots__ = ("_processor", "_name", "_source", "_destinations", "_value")

    def __init__(self, processor: Processor, name: str = "Signal", value=None):
        self._processor = processor
        self._name = name
        self._source = None
        self._destinations = set()
        self._value = value

    def add_destination(self, module: Module) -> None:
        """Add a module as a destination that the signal will send new
        values to.
        """
        self._destinations.add(module)

    def update_value(self, value) -> None:
        """
        Set the value of the signal and queue destination modules for update.

        Parameters
        ----------
        value : Any
            Value to set the signal to.
        """
        if value != self._value:
            self._value = value
            # Inform the processor on what modules have been given
            # signal so they can update
            for destination in self._destinations:
                self._processor.add_modules_to_update(destination)

    def get_value(self) -> Any:
        """Get the current value of the signal.

        Returns
        -------
        Any
            Value of the signal.
        """
        return self._value

    def set_value(self, value: Any) -> None:
        """Set the value of the signal without queuing the destination
        modules of the signal for update.

        This method should not be called by a module.

        Parameters
        ----------
        value : any
            Value to set the signal to.
        """
        self._value = value

    def get_name(self) -> str:
        """Return the name of the signal.

        Returns
        -------
        str
            Name of the signal.
        """
        return self._name
