import ast
import json
import sys
from json import JSONDecodeError
from threading import Thread

from qtpy import QtCore, QtWidgets
from qtpy.QtCore import QPointF
from qtpy.QtCore import Signal as pyqtSignal
from qtpy.QtCore import Slot
from qtpy.QtWidgets import (
    QAction,
    QApplication,
    QErrorMessage,
    QFileDialog,
    QGraphicsView,
    QInputDialog,
    QLabel,
    QMainWindow,
    QMessageBox,
    QSpinBox,
    QStyle,
    QWidget,
)

from simudator.core.processor import Processor
from simudator.gui.breakpoint_window import BreakpointWindow
from simudator.gui.cpu_graphics_scene import CpuGraphicsScene
from simudator.gui.custom_toolbar import CustomToolBar
from simudator.gui.dialogs.lambda_breakpoint_dialog import LambdaBreakpointDialog
from simudator.gui.module_graphics_item.module_graphics_item import ModuleGraphicsItem
from simudator.gui.orientation import Orientation
from simudator.gui.port_graphics_item import PortGraphicsItem
from simudator.gui.run_continuously_thread import RunThread
from simudator.gui.signal_graphics_item import SignalGraphicsItem
from simudator.gui.pipeline import PipeLine


class View(QGraphicsView):
    """
    This class views all QGraphicsItems for the user. It takes the
    QGraphicsScene as input and inherits all funcitonality from the
    QGraphicsView and overrides the wheelEvent function.
    This allows the users to navigate the view with their trackpads
    and zoom in/out with their trackpads + ctrl.
    """

    def __init__(self, QGraphicsScene):
        super().__init__(QGraphicsScene)
        self.scene = QGraphicsScene

    def wheelEvent(self, event):
        """
        Default behaviour if ctrl is not pressed, otherwise zoom in/out.
        """
        modifiers = QtWidgets.QApplication.keyboardModifiers()
        if modifiers == QtCore.Qt.ControlModifier:
            # Factor above 1 zooms in, below zooms out
            factor = 1.03
            if event.angleDelta().y() < 0:
                factor = 0.97

            # If event got triggered due to the x axis, do nothing
            if event.pixelDelta().x() != 0:
                return

            view_pos = event.globalPosition()
            scene_pos = self.mapToScene(int(view_pos.x()), int(view_pos.y()))

            self.centerOn(scene_pos)
            self.scale(factor, factor)

            old_mapToScene = self.mapToScene(int(view_pos.x()), int(view_pos.y()))
            new_mapToScene = self.mapToScene(self.viewport().rect().center())

            delta = old_mapToScene - new_mapToScene

            self.centerOn(scene_pos - delta)

        else:
            # Default behaviour
            super().wheelEvent(event)


class GUI(QMainWindow):
    """
    Main gui class. Handles gui windows, toolbar and visualizes modules.

    This is the main class for the GUI. It handles creating the window for
    the gui, aswell as the toolbar for controlling the simultaion.
    It takes a processor and visualizes its modules as boxes with the signals as lines
    between them. Graphics items for the modules and signals need to be created
    and added individually.
    """

    cpu_tick_signal = pyqtSignal(int)
    HALT_MESSAGE_THRESHOLD = 100

    def __init__(self, cpu: Processor):
        super().__init__()

        self.cpu = cpu

        self.setWindowTitle("SimuDator")
        self.cpu_graphics_scene = CpuGraphicsScene(cpu)
        self.graphics_view = View(self.cpu_graphics_scene)
        # self.graphics_view.setDragMode(True)
        self.moduleActions: dict[str, QAction] = {}

        self.setCentralWidget(self.graphics_view)

        self.errorMessageWidget = QErrorMessage(self)

        self.createToolBar()

        # Threadpool for running cpu and gui concurently
        self.threadpool = QtCore.QThreadPool()

        # Signal to tell gui when cpu has halted
        self.cpu_tick_signal.connect(self.handleCpuTick)

        # Used to set if all values in the gui should be updated each tick
        # or only the clock counter
        self.update_all_values = False

        # Used to set the update delay
        # Useful when watching the values being updated while running
        self.update_delay = 0.00

        # Used to lock some actions in ui when cpu is running in another thread
        # Using the cpu's internal status directly could case problems
        self.cpu_running = False

        self.pipeline = PipeLine()
        self.init_pipeline()

        # Set Style, THESE ARE TEST AND DONT WORK
        app = QApplication.instance()
        app.setStyleSheet("Qwidget.QMainWindow { background-color: yellow }")
        self.setAttribute(QtCore.Qt.WA_StyledBackground)

        # Add this so we can save the window in scope later, otherwise it disappears
        self.breakpoint_window = None

    def createToolBar(self) -> None:
        """
        Creates the toolbar containing file, layout and toolbar buttons.


        Creates the toolbar containing the file, layout and toolbar
        buttons with their respective sub buttons.
        """

        # Create toolbar
        toolbar = CustomToolBar("Main toolbar")
        self.addToolBar(toolbar)

        # Create load action
        self.load_action = QAction("Load", self)
        self.load_action.setStatusTip("Load processor state from file")
        self.load_action.triggered.connect(self.loadToolBarButtonClick)

        # Create save action
        self.save_action = QAction("Save", self)
        self.save_action.setStatusTip("Save processor state to file")
        self.save_action.triggered.connect(self.saveStateMenuButtonClick)

        self.reset_action = QAction("Reset", self)
        self.reset_action.setStatusTip("Reset processor")
        self.reset_action.triggered.connect(self.resetMenuButtonClick)

        # Create File menu for load and save
        menu = self.menuBar()
        file_menu = menu.addMenu("&File")
        file_menu.addAction(self.load_action)
        file_menu.addAction(self.save_action)
        file_menu.addAction(self.reset_action)

        # create load layout action
        load_layout_action = QAction("Load layout", self)
        load_layout_action.setStatusTip("Loads the selected layout from file.")
        load_layout_action.triggered.connect(self.loadLayoutToolBarButtonClick)

        # create load default layout action
        load_default_layout_action = QAction("Load default layout", self)
        load_default_layout_action.setStatusTip("Loads the default layout from file.")
        load_default_layout_action.triggered.connect(
            self.loadDefaultLayoutToolBarButtonClick
        )

        # create save layout action
        save_layout_action = QAction("Save layout", self)
        save_layout_action.setStatusTip("Saves the current layout to a file.")
        save_layout_action.triggered.connect(self.saveLayoutToolBarButtonClick)

        # create layout lock action
        self.lock_layout_action = QAction("Lock layout", self, checkable=True)
        self.lock_layout_action.setStatusTip("Lock layout so items can't be moved.")
        self.lock_layout_action.triggered.connect(self.toggleLayoutLockMenuButtonClick)

        # Create show signals actions
        self.signal_vis_action = QAction("Show signals", self, checkable=True)
        self.signal_vis_action.setChecked(True)
        self.signal_vis_action.setStatusTip("Toggle the visibility of signal.")
        self.signal_vis_action.triggered.connect(self.showSignalsMenuButtonClick)

        # Create show port name actions
        self.port_vis_action = QAction("Show port names", self, checkable=True)
        self.port_vis_action.setChecked(True)
        self.port_vis_action.setStatusTip("Toggle the visibility of port names.")
        self.port_vis_action.triggered.connect(self.showPortNamesBarButtonClick)

        # Create Layout menu for layout actions
        layout_menu = menu.addMenu("&Layout")
        layout_menu.addAction(load_layout_action)
        layout_menu.addAction(load_default_layout_action)
        layout_menu.addAction(save_layout_action)
        layout_menu.addAction(self.lock_layout_action)
        layout_menu.addAction(self.signal_vis_action)
        layout_menu.addAction(self.port_vis_action)

        # Create breakpoint window action
        self.breakpoint_action = QAction("Breakpoints", self)
        self.breakpoint_action.setStatusTip("Open breakpoint window.")
        self.breakpoint_action.triggered.connect(self.openBreakpointWindow)

        # Create 'update value' window button
        self.update_value_action = QAction(
            "Update values while running", self, checkable=True
        )
        self.update_value_action.setChecked(False)
        self.update_value_action.setStatusTip("Toggle value updates while running.")
        self.update_value_action.triggered.connect(self.toggle_value_update_on_run)

        # Create 'set delay' window button
        self.set_delay_action = QAction("Set update delay", self)
        self.set_delay_action.setStatusTip(
            "Sets the delay between each update when the cpu is running."
        )
        self.set_delay_action.triggered.connect(self.set_update_delay)

        # Create Tools menu for tool actions actions
        tools_menu = menu.addMenu("&Tools")
        tools_menu.addAction(self.breakpoint_action)
        tools_menu.addAction(self.update_value_action)
        tools_menu.addAction(self.set_delay_action)

        # Add run button on toolbar
        arrow_icon = self.style().standardIcon(QStyle.SP_MediaPlay)
        self.run_action = QAction(arrow_icon, "Run", self)
        self.run_action.setStatusTip("Run until halt")
        self.run_action.triggered.connect(self.runToolBarButtonClick)
        toolbar.addAction(self.run_action)

        # Add stop button on toolbar
        stop_icon = self.style().standardIcon(QStyle.SP_MediaStop)
        self.stop_action = QAction(stop_icon, "Stop", self)
        self.stop_action.setStatusTip("Stop running")
        self.stop_action.triggered.connect(self.stopToolBarButtonClick)
        toolbar.addAction(self.stop_action)

        # Add Asm label
        self.clock_cycle_label = QLabel("Clock cycle: ", self)
        toolbar.addWidget(self.clock_cycle_label)

        # Add undo button on toolbar
        backward_arrow_icon = self.style().standardIcon(QStyle.SP_MediaSeekBackward)
        self.undo_action = QAction(backward_arrow_icon, "Undo", self)
        self.undo_action.setStatusTip("Undo the last processor tick")
        self.undo_action.triggered.connect(self.undoToolBarButtonClick)
        toolbar.addAction(self.undo_action)

        # Add step button on toolbar
        forward_arrow_icon = self.style().standardIcon(QStyle.SP_MediaSeekForward)
        self.step_action = QAction(forward_arrow_icon, "Step", self)
        self.step_action.setStatusTip("Run one clock cycle")
        self.step_action.triggered.connect(self.stepToolBarButtonClick)
        toolbar.addAction(self.step_action)

        # Add box for jump value
        self.jump_value_box = QSpinBox()
        self.jump_value_box.setMinimum(999999)
        self.jump_value_box.setMinimum(1)
        self.jump_value_box.setValue(1)
        toolbar.addWidget(self.jump_value_box)

        # Add seperator so clock gets better spacing
        spacing = QWidget()
        spacing.setFixedWidth(10)
        toolbar.addWidget(spacing)
        toolbar.addSeparator()
        spacing = QWidget()
        spacing.setFixedWidth(10)
        toolbar.addWidget(spacing)

        # Add clock counter
        self.clock_label = QLabel("Clock cycles: " + str(self.cpu.get_clock()), self)
        toolbar.addWidget(self.clock_label)

        # Add seperator so clock gets better spacing
        spacing = QWidget()
        spacing.setFixedWidth(10)
        toolbar.addWidget(spacing)
        toolbar.addSeparator()
        spacing = QWidget()
        spacing.setFixedWidth(10)
        toolbar.addWidget(spacing)

        # Add Asm label
        self.asm_label = QLabel("Assembler Instructions: ", self)
        toolbar.addWidget(self.asm_label)

        # Add undo asm button on toolbar
        backward_arrow_icon = self.style().standardIcon(QStyle.SP_MediaSeekBackward)
        self.undo_asm_action = QAction(backward_arrow_icon, "Undo Asm", self)
        self.undo_asm_action.setStatusTip("Undo the last assembler instruction")
        self.undo_asm_action.triggered.connect(self.undoAsmToolBarButtonClick)
        toolbar.addAction(self.undo_asm_action)

        # Add step button on toolbar
        forward_arrow_icon = self.style().standardIcon(QStyle.SP_MediaSeekForward)
        self.step_asm_action = QAction(forward_arrow_icon, "Step Asm", self)
        self.step_asm_action.setStatusTip("Run one assembler instruction")
        self.step_asm_action.triggered.connect(self.stepAsmToolBarButtonClick)
        toolbar.addAction(self.step_asm_action)

        # Add box for jump value
        self.asm_jump_value_box = QSpinBox()
        self.asm_jump_value_box.setMinimum(999999)
        self.asm_jump_value_box.setMinimum(1)
        self.asm_jump_value_box.setValue(1)
        toolbar.addWidget(self.asm_jump_value_box)

    def connectModuleActions(self, action_signals: []) -> None:
        """
        Connects modules with actions thorught ids (strings). Useful for breakpoints.

        This function takes a list of tuples with an
        action id (which is a predefined string) and a signal.
        It then connects the signal to a function depending on the action id.
        We do this so graphics items actions can access functionalty
        that is contolled by the gui and cpu. Mainly creating breakpoints.

        We do this instead of sending the graphics items the cpu or
        creating actions in the gpu. Sending the cpu directly to the
        graphics items could become problematic when many sources
        affect the cpu in unexpected ways, and creating these
        actions in the gpu would make the graphics items much less modular.
        """
        for action_id, signal in action_signals:
            match action_id:
                case "BP":
                    signal.connect(self.addStateBreakpoint)
                case "MEM_BP":
                    signal.connect(self.addMemoryBreakpoint)
                case "LMB_BP":
                    # LambdaBPs are not tested since we dont
                    # actually have any modules that make them yet
                    signal.connect(self.addLambdaBreakpoint)
                case "UPDATE":
                    signal.connect(self.updateCpuListeners)
                case "CLOCKUPDATE":
                    signal.connect(self.updateCpuClockCycle)

    def init_pipeline(self) -> None:
        """Initialize the pipeline diagram. 

        Sets its height, width and instructions.
        """
        size = self.cpu.get_pipeline_dimensions()

        self.pipeline.set_height(size[0])
        self.pipeline.set_width(size[1])
        self.update_pipeline()

        self.pipeline.show()

    def updateCpuListeners(self) -> None:
        """
        Updates the graphics items in the scene, the clock, and the pipeline diagram.

        Used after the cpu has run or when the user has edited somehting.
        """
        self.cpu_graphics_scene.updateGraphicsItems()
        self.updateCpuClockCycle()
        self.update_pipeline()

    def update_pipeline(self) -> None:
        self.pipeline.set_instructions(self.cpu.get_current_instructions())

    def updateCpuClockCycle(self) -> None:
        """
        Update the clock cycle counter.

        Used while the program is running to show the user nothing has crashed.
        """
        self.clock_label.setText("Clockcycle: " + str(self.cpu.get_clock()))

    def lambdaBreakpointDialog(self) -> None:
        """
        Opens dialog window for user to create a breakpoint.
        """
        lambdas = self.cpu.get_breakpoint_lambdas()
        lambda_br_dialog = LambdaBreakpointDialog(lambdas, self)
        lambda_br_dialog.accepted.connect(self.addLambdaBreakpoint)

    def updateBreakpointWindow(self) -> None:
        """
        Updates the breakpoint window when new breakpoints are added.
        """
        # Don't do anything if window is closed
        if self.breakpoint_window is not None:
            self.breakpoint_window.update()

    """
    @Slot is used to explicitly mark a python method as a Qt slot
    and specify a C++ signature for it, which is used most commonly
    in order to select a particular overload.
    """

    @Slot(str, str, str)
    def editModuleState(self, module_name, state, value) -> None:
        """
        Tries to change the value of module to value given by user.

        Does this through the dialog opened by editModuleStateDialog.
        """
        try:
            parsed_value = ast.literal_eval(value)
        except SyntaxError as e:
            self.errorMessageWidget.showMessage(str(e))
        else:
            module = self.cpu.get_module(module_name)
            module_state = module.get_state()
            module_state[state] = parsed_value
            module.set_state(module_state)
            self.cpu_graphics_scene.updateGraphicsItems()

    @Slot(str, str, str)
    def editMemoryContent(self, module_name: str, adress: str, value: str) -> None:
        """
        Tries to change the value at adress to value given by user.

        Does this through the dialog opened by editModuleStateDialog.
        """
        try:
            parsed_adress = int(adress, 16)
            parsed_value = ast.literal_eval(value)
        except SyntaxError as e:
            self.errorMessageWidget.showMessage(str(e))
        except ValueError:
            self.errorMessageWidget.showMessage(
                "You must enter a hexadecimal" "number preceeded by '0x' (e.g." "0xc3)."
            )
        else:
            module = self.cpu.get_module(module_name)
            module_state = module.get_state()
            module_state['memory'][parsed_adress] = parsed_value
            module.set_state(module_state)
            self.cpu_graphics_scene.updateGraphicsItems()

    @Slot(str, str, str)
    def addStateBreakpoint(self, module_name: str, state: str, value: str) -> None:
        """
        Tries to add a breakpoint to the module through the dialog opened by stateBreakpointDialog.
        """
        try:
            parsed_value = ast.literal_eval(value)
            self.cpu.add_state_breakpoint(module_name, state, parsed_value)
            self.updateBreakpointWindow()
        except (ValueError, SyntaxError) as e:
            self.errorMessageWidget.showMessage(str(e))

    @Slot(str, str)
    def addLambdaBreakpoint(self, lambda_name, kwargs_str) -> None:
        """
        Tries to add a breakpoint to the module through the dialog opened by lambdaBreakpointDialog.
        """
        try:
            lambda_kwargs = {}
            for kwarg in kwargs_str.split(' '):
                key, value = kwarg.split('=')
                lambda_kwargs[key] = ast.literal_eval(value)
            self.cpu.add_lambda_breakpoint(lambda_name, **lambda_kwargs)
            self.updateBreakpointWindow()
        except (ValueError, SyntaxError) as e:
            self.errorMessageWidget.showMessage(str(e))

    @Slot(str, str, str)
    def addMemoryBreakpoint(self, module_name: str, adress: str, value: str) -> None:
        """
        Tries to add a breakpoint to the module through the dialog opened by memoryBreakpointDialog.
        """
        try:
            parsed_adress = int(adress)
            parsed_value = int(value)
            self.cpu.add_memory_breakpoint(module_name, parsed_adress, parsed_value)
            self.updateBreakpointWindow()
        except SyntaxError as e:
            self.errorMessageWidget.showMessage(str(e))

    def setDisabledWhenRunning(self, is_disable):
        """
        Greys out buttons for actions that can't be done while cpu is running.
        """
        self.load_action.setDisabled(is_disable)
        self.save_action.setDisabled(is_disable)
        self.reset_action.setDisabled(is_disable)
        self.run_action.setDisabled(is_disable)
        self.step_action.setDisabled(is_disable)
        self.undo_action.setDisabled(is_disable)

    def stepToolBarButtonClick(self):
        """
        Runs the cpu a specified number of clock cycles according to the jump value box.
        """

        # Don't do steps if cpu is running
        if self.cpu_running:
            return

        steps = self.jump_value_box.value()
        self.cpu_running = True
        self.setDisabledWhenRunning(True)
        simulation_thread = RunThread(
            self.cpu, self.cpu_tick_signal, self.update_delay, False, False, steps
        )
        self.threadpool.start(simulation_thread)

    def stepAsmToolBarButtonClick(self):
        """
        Runs the cpu a specified number of asm instructions according to the jump value box.
        """

        # Don't do steps if cpu is running
        if self.cpu_running:
            return

        steps = self.asm_jump_value_box.value()
        self.cpu_running = True
        self.setDisabledWhenRunning(True)
        simultaion_thread = RunThread(
            self.cpu, self.cpu_tick_signal, self.update_delay, False, True, steps
        )
        self.threadpool.start(simultaion_thread)
        self.updateCpuListeners()

    def runToolBarButtonClick(self) -> None:
        """
        Runs the cpu until it is stopped by user input, breakpoint or similar.
        """

        # Don't run if already running
        if self.cpu_running:
            return

        # Create own thread for cpu simulation so gui dosent freeze
        self.cpu_running = True
        self.setDisabledWhenRunning(True)
        simulation_thread = RunThread(self.cpu, self.cpu_tick_signal, self.update_delay)
        self.threadpool.start(simulation_thread)

    @Slot(int)
    def handleCpuTick(self, steps: int) -> None:
        """
        Called from other thread after every cpu tick.
        Will inform the user and update visuals.
        """

        # Update cpu clock counter every tick
        self.updateCpuClockCycle()

        if self.update_all_values:
            self.updateCpuListeners()

        # A signal of 0 steps signifies end of execution, i.e. the CPU has
        # halted or run the specified amount of ticks
        # => Enable the relevant parts of the GUI again
        if steps == 0:
            self.cpu_running = False
            self.setDisabledWhenRunning(False)
            self.updateCpuListeners()

            # Inform user of reached break point
            if self.cpu.breakpoint_reached:
                self.messageBox(
                    "Reached breakpoint: " + self.cpu.last_breakpoint.__str__()
                )

            # Inform user of halt
            if self.cpu.should_halt():
                self.messageBox("The processor halted.")

    def stopToolBarButtonClick(self) -> None:
        """
        Tells the cpu to stop. It will then stop at an appropriate in its own thread.
        """
        self.cpu.stop()

    def folderSaveDialog(self) -> str:
        """
        Open a file explorer in a new window. Return the absolute path to the selected file.

        Can return existing as well as non-existing files.
        If the selected files does not exist it will be created.
        """
        dialog = QFileDialog()
        dialog.setFileMode(QFileDialog.AnyFile)
        dialog.setAcceptMode(QFileDialog.AcceptOpen)
        dialog.setDirectory("~/simudator")  # TODO: does this work when exported?

        # getSaveFileName() can returncan return existing file but also create new ones
        return dialog.getSaveFileName()[0]

    def folderLoadDialog(self) -> str:
        """
        Open a file explorer in a new window. Return the absolute path to the selected file.

        Can only return existing files.
        """
        dialog = QFileDialog()
        dialog.setFileMode(QFileDialog.AnyFile)
        dialog.setAcceptMode(QFileDialog.AcceptOpen)
        dialog.setDirectory("~/simudator")  # TODO: does this work when exported?

        # getOpenFileName() will only return already existing files
        return dialog.getOpenFileName()[0]

    def loadToolBarButtonClick(self) -> None:
        """
        Loads the processor state from selected file.
        """

        # not safe to load while cpu is running
        if self.cpu_running:
            return

        path = self.folderLoadDialog()
        try:
            self.cpu.load_state_from_file(path)
        except KeyError:
            self.errorBox("Could not load selected file.")
        except ValueError:
            self.errorBox("Selected file was empty.")
        except FileNotFoundError:
            # This error is triggered when no file was selected by
            # by the user. The user should know they did not select a
            # file and therefore there no need to display a box telling
            # them that.
            pass
            # self.errorBox("No vaild file was selected.")
        else:
            self.messageBox("Loaded file successfully.")
            self.updateCpuListeners()

    def resetMenuButtonClick(self) -> None:
        """
        Will reset processor to inital values.
        """

        # not safe to reset while cpu running
        if self.cpu_running:
            return

        answer = QMessageBox.question(
            self, "Reset Processor", "Are you sure you want to reset the processor?"
        )

        if answer == QMessageBox.Yes:
            self.cpu.reset()
            self.updateCpuListeners()

    def undoNToolBarButtonClick(self) -> None:
        # TODO: I dont think this is used
        """
        Undo zero to N cycles.
        """

        # Don't try to undo while running
        if self.cpu_running:
            return

        cycles, ok = QInputDialog(self).getInt(
            self,
            "Input number of cycles to run",
            "Input number of cycles to run",
        )
        if ok:
            if cycles < 1:
                self.errorBox("Please input a number larger than 0.")
                return
            try:
                # Make sure we load 0 as lowest
                self.cpu.load_cycle(max(self.cpu.get_clock() - cycles, 0))
                self.updateCpuListeners()
            except (ValueError, IndexError):
                self.errorBox("Unable to undo the cycle(s).")

    def saveLayoutToolBarButtonClick(self) -> None:
        """
        Saves the layout of all the modules in a (somewhat) human readable format.

        This also erases the previous content of the file before saving
        the data.
        """
        path = self.folderSaveDialog()

        # Change file to the selected file if a file was selected
        if path == '':
            return

        # Open the given file erasing the previous content
        with open(path, "w") as fp:
            graphics_modules_data = {}
            ports_data = {}
            graphics_signals_data = {}

            graphics_modules = self.cpu_graphics_scene.getModulesGraphicsItems()
            for graphics_module in graphics_modules:
                pos = (graphics_module.x(), graphics_module.y())
                graphics_modules_data[graphics_module.getName()] = pos

                for port in graphics_module.getPorts():
                    visibility = port.isVisible()
                    orientation = int(port.getOrientation())
                    data = (port.x(), port.y(), orientation, visibility)
                    ports_data[port.getID()] = data

            for graphics_signal in self.cpu_graphics_scene.getSignalGraphicsItems():
                visibility = graphics_signal.isVisible()
                points = []
                for point in graphics_signal.getPoints():
                    points.append((point.x(), point.y()))
                data = (points, visibility)
                graphics_signals_data[graphics_signal.getID()] = data

            data = (graphics_modules_data, ports_data, graphics_signals_data)
            json.dump(data, fp)

        fp.close()

    def loadDefaultLayoutToolBarButtonClick(self) -> None:
        """
        Places all the module_graphic objects in a diagonal going down to the right.
        """
        counter = 0
        for key in self.cpu_graphics_scene.module_graphics_items:
            module_graphic = self.cpu_graphics_scene.module_graphics_items[key]
            module_graphic.showPorts()
            counter += 1
            self.cpu_graphics_scene.replaceModuleGraphicsItem(
                module_graphic, counter, counter
            )
        self.cpu_graphics_scene.resetSignals()

    def loadLayoutFromFile(self, file) -> None:
        """
        Loads a layout for the current cpu from the selected file.

        If at anypoint this funciton would error, the default layout
        is loaded instead.
        """
        # TODO: speed(?)
        graphics_modules = self.cpu_graphics_scene.moduleGraphicsItemsDict()
        ports = {}
        graphics_signals = {}

        for graphics_module in graphics_modules.values():
            for port in graphics_module.getPorts():
                ports[port.getID()] = port

        for graphics_signal in self.cpu_graphics_scene.getSignalGraphicsItems():
            graphics_signals[graphics_signal.getID()] = graphics_signal

        # Open the file in 'read-only'
        with open(file, 'rb') as fp:
            data = json.load(fp)
            graphics_modules_data = data[0]
            ports_data = data[1]
            graphics_signals_data = data[2]

            for g_module_name, g_module_data in graphics_modules_data.items():
                g_module = graphics_modules[g_module_name]
                x = g_module_data[0]
                y = g_module_data[1]
                self.loadGraphicsModule(g_module, x, y)

            for port_id, port_data in ports_data.items():
                port = ports[int(port_id)]
                self.loadPort(port, *port_data)

            for g_signal_id, g_signal_data in graphics_signals_data.items():
                g_signal = graphics_signals[int(g_signal_id)]
                self.loadSignal(g_signal, *g_signal_data)

        fp.close()

    def loadSignal(
        self,
        graphic_signal: SignalGraphicsItem,
        signal_points: list[tuple[float, float]],
        visibility: bool,
    ) -> None:
        """
        Changes the graphical signal to have the positions given as argument.
        """
        qpoints = []
        # Turn points -> QPointF
        # list[int, int] -> QPointF object
        for point in signal_points:
            qpoints.append(QPointF(point[0], point[1]))

        # Set the new points
        graphic_signal.setPoints(qpoints)
        graphic_signal.setVisible(visibility)

    def loadGraphicsModule(
        self,
        graphics_module: ModuleGraphicsItem,
        graphics_module_x: float,
        graphics_module_y: float,
    ) -> None:
        """
        Changes the positions of graphical modules to the ones given as argument.
        """
        graphics_module.setX(graphics_module_x)
        graphics_module.setY(graphics_module_y)

    def loadPort(
        self,
        port: PortGraphicsItem,
        x: float,
        y: float,
        orientation: Orientation,
        visibility: bool,
    ) -> None:
        port.setOrientation(orientation)
        port.setX(x)
        port.setY(y)
        port.setVisible(visibility)

    def saveStateMenuButtonClick(self) -> None:
        """
        Save state of cpu to file.

        This erases the previous content of the file before saving the data.
        """

        # Not safe to save while running
        if self.cpu_running:
            return

        path = self.folderSaveDialog()

        # Change file to the selected file if a file was selected
        if path == '':
            return

        self.cpu.save_state_to_file(path)

    def openBreakpointWindow(self) -> None:
        """
        Opens window for editing breakpoints.
        """
        self.breakpoint_window = BreakpointWindow(self.cpu)
        self.breakpoint_window.show()

    def toggle_value_update_on_run(self):
        """
        Toggles whether all values or only clock cycle is being updated each tick.
        """
        self.update_all_values = not self.update_all_values

    def set_update_delay(self):
        """
        Sets the update delay for the visual updates while the cpu is running.
        """
        delay, ok = QInputDialog.getDouble(
            self, "Input Dialog", "Enter a float value:", decimals=5
        )
        if ok:
            self.update_delay = delay

    def showPortNamesBarButtonClick(self):
        """
        Toggles showing port names in the graphics scene.
        """
        self.cpu_graphics_scene.setPortNamesVisibility(self.port_vis_action.isChecked())

    def showSignalsMenuButtonClick(self) -> None:
        """
        Toggle shoing the signals in the graphics scene.
        """
        self.cpu_graphics_scene.setAllSignalsVisibility(
            self.signal_vis_action.isChecked()
        )

    def toggleLayoutLockMenuButtonClick(self) -> None:
        """
        Toggles so the layout can not be edited.
        """
        self.cpu_graphics_scene.setLayoutLock(self.lock_layout_action.isChecked())

    def loadLayoutToolBarButtonClick(self) -> None:
        """
        Loads a given layout from a selected file.

        If the layout was
        unable to load, an error message will pop up informing the user
        and the default layout will be loaded.
        """

        path = self.folderLoadDialog()

        # If no file was selected, do nothing
        if path == '':
            return

        try:
            self.loadLayoutFromFile(path)

        # Anything goes wrong with loading the selected one,
        # we dont care about what went wrong
        except (OSError, JSONDecodeError):
            self.loadDefaultLayoutToolBarButtonClick()
            self.errorBox("Unable to load given file.")
        else:
            self.messageBox("File loaded successfully.")

    def errorBox(self, message="Something went wrong.") -> None:
        """
        Displays a simple box with the given error message.
        """

        dlg = QMessageBox(self)
        dlg.setWindowTitle("Error")
        dlg.setText(message)
        dlg.exec()

    def messageBox(self, message="Something happend.") -> None:
        """
        Displays a simple box with the given message.
        """

        dlg = QMessageBox(self)
        dlg.setWindowTitle("Message")
        dlg.setText(message)
        dlg.exec()

    def undoToolBarButtonClick(self) -> None:
        """
        Undos as many processor cycles as the number entered in the box.
        """
        try:
            steps = self.jump_value_box.value()
            self.cpu.load_cycle(max(self.cpu.get_clock() - steps, 0))
            self.updateCpuListeners()
        except (ValueError, IndexError):
            self.errorBox("Unable to undo the cycle.")

    def undoAsmToolBarButtonClick(self) -> None:
        """
        Undos as many processor cycles as the number entered in the box.
        """
        try:
            steps = self.asm_jump_value_box.value()
            self.cpu.undo_asm_instruction(steps)
            self.updateCpuListeners()
        except (ValueError, IndexError):
            self.errorBox("Unable to undo the cycle.")

    def addModuleGraphicsItem(self, graphics_item: ModuleGraphicsItem) -> None:
        """
        Adds an item to the graphics scene.
        """
        self.cpu_graphics_scene.addModuleGraphicsItem(graphics_item)
        self.connectModuleActions(graphics_item.getActionSignals())

    def addAllSignals(self) -> None:
        """
        Add signals depending on modules in the graphics scene.
        """
        self.cpu_graphics_scene.addAllSignals()


if __name__ == '__main__':
    QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling)
    app = QApplication(sys.argv)
    gui = GUI(Processor())
    gui.show()
    app.exec()
