import itertools
from typing import Optional

from qtpy.QtCore import QObject, QPointF, Qt
from qtpy.QtCore import Signal as pyqtSignal
from qtpy.QtWidgets import (
    QAction,
    QGraphicsItemGroup,
    QGraphicsLineItem,
    QGraphicsSceneMouseEvent,
    QGraphicsSimpleTextItem,
    QMenu,
)

from simudator.core.signal import Signal
from simudator.gui.color_scheme import ColorScheme
from simudator.gui.orientation import Orientation


class PyQtSignalHolder(QObject):
    """
    A helper class for PortGraphicsItem that stores pyqtSignal instances.
    This is needed because subclasses of QGraphicsItemGroup cannot inherit
    from QObject or QGraphicsObject to have pyqtSignals like normally
    done for QGraphicsItem subclasses that need pyqtSignals.
    """

    moved = pyqtSignal(name="moved")
    toggled = pyqtSignal(name="toggled")

    def __init__(self, parent: Optional['QObject'] = None) -> None:
        super().__init__(parent)


class PortGraphicsItem(QGraphicsItemGroup):
    """
    Graphics item that visualises the connection between a module and a signal.
    Does this by drawing a small line that goes out from the
    module and the name of the signal. The port can be moved around
    the edge of the of the module grapics item.
    """

    LINE_LENGTH = 8
    CHAR_HEIGHT = 18
    CHAR_LEN = 6
    FONT_SIZE = 8
    id_iter = itertools.count()

    def __init__(
        self,
        signal: Signal,
        orientation: Orientation = Orientation.RIGHT,
        parent=None,
        display_name: str = None,
    ):
        super().__init__(parent)
        # Use the name of the signal if no display name is given
        if display_name is None:
            self.name = signal.get_name()
        else:
            self.name = display_name

        self.id = next(self.id_iter)
        self.parent = parent
        self.signal = signal
        self.orientation = orientation
        self.isLocked = False

        # Set up pyqt signals using a QObject instance since inheriting from
        # QObject or QGraphicsObject directly does not work with subclasses
        # of QGraphicsItemGroup
        self.pyqtSignalHolder = PyQtSignalHolder()
        self.moved = self.pyqtSignalHolder.moved
        self.toggled = self.pyqtSignalHolder.toggled

        self.drawPort()
        self.setOrientation(orientation)
        self.left_click = False
        self.right_click = False

        # Draw ports second, it recivies mouse events over the signals
        self.setZValue(1.0)

    def drawPort(self) -> None:
        """
        Draws the port as a small line and a label with the signals name.
        """

        # Create a horizontal and vertical line
        # We create two so we dont have to delete them when changing orientation
        # and can instead hide one
        self.vertical_line = QGraphicsLineItem(0, 0, 0, self.LINE_LENGTH)
        self.horizontal_line = QGraphicsLineItem(0, 0, self.LINE_LENGTH, 0)
        self.name_label = QGraphicsSimpleTextItem(self.name)

        # Set fontsize
        font = self.name_label.font()
        font.setPointSize(self.FONT_SIZE)
        self.name_label.setFont(font)

        # Set colors
        self.vertical_line.setPen(ColorScheme.Port)
        self.horizontal_line.setPen(ColorScheme.Port)
        self.name_label.setBrush(ColorScheme.PortLabel)

        self.addToGroup(self.vertical_line)
        self.addToGroup(self.horizontal_line)
        self.addToGroup(self.name_label)

    def setOrientation(self, orientation: int) -> None:
        """
        Sets the ports orientation by showing the vertical or horizontal line
        and moving the name label appropriatly.
        """
        self.orientation = orientation
        if orientation == Orientation.UP:
            # Set right line visible
            self.vertical_line.setVisible(True)
            self.horizontal_line.setVisible(False)

            # set line position
            self.vertical_line.setPos(0, -self.LINE_LENGTH)

            # set label position
            self.name_label.setPos(
                -len(self.name) * self.CHAR_LEN / 2, -self.LINE_LENGTH * 3.2
            )

        elif orientation == Orientation.DOWN:
            # Set right line visible
            self.vertical_line.setVisible(True)
            self.horizontal_line.setVisible(False)

            # set line position
            self.vertical_line.setPos(0, 0)

            # set label position
            self.name_label.setPos(
                -len(self.name) * self.CHAR_LEN / 2, self.LINE_LENGTH * 1.2
            )

        elif orientation == Orientation.LEFT:
            # Set right line visible
            self.vertical_line.setVisible(False)
            self.horizontal_line.setVisible(True)

            # set line position
            self.horizontal_line.setPos(-self.LINE_LENGTH, 0)

            # set label position
            self.name_label.setPos(
                -self.LINE_LENGTH * 1.5 - len(self.name) * self.CHAR_LEN,
                -self.CHAR_HEIGHT,
            )

        elif orientation == Orientation.RIGHT:
            # Set right line visible
            self.vertical_line.setVisible(False)
            self.horizontal_line.setVisible(True)

            # set line position
            self.horizontal_line.setPos(0, 0)

            # set label position
            self.name_label.setPos(self.LINE_LENGTH * 1.5, -self.CHAR_HEIGHT)

    def getPoint(self) -> QPointF:
        """
        Gives the position of the "point" of the port line graphic.
        Used by signal graphics to know where ot draw.
        """
        line_length = QPointF(0, 0)

        # We need to compensate with line lenght depending on orientation
        if self.orientation == Orientation.UP:
            line_length = QPointF(0, -self.LINE_LENGTH)
        elif self.orientation == Orientation.DOWN:
            line_length = QPointF(0, self.LINE_LENGTH)
        elif self.orientation == Orientation.LEFT:
            line_length = QPointF(-self.LINE_LENGTH, 0)
        elif self.orientation == Orientation.RIGHT:
            line_length = QPointF(self.LINE_LENGTH, 0)

        return self.scenePos() + line_length

    def mousePressEvent(self, event):
        super().mousePressEvent(event)

        # If event is not accepted it will be triggered twice
        if event.button() == Qt.MouseButton.LeftButton:
            self.left_click = True

        event.accept()

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        """
        Moves the port if left click is held. Will move it around
        the edge of it parent modules baserect.
        """
        if not self.left_click and self.parent is not None:
            return
        if self.isLocked:
            return

        # Get mouses relative position to parent
        current_mouse_point = event.scenePos()
        relative_mouse_pos_x = current_mouse_point.x() - self.parent.x()
        relative_mouse_pos_y = current_mouse_point.y() - self.parent.y()

        # Get min and max positions depending on the parents base rectangle
        bounding_rect = self.parent.baserect
        min_x = bounding_rect.x()
        max_x = bounding_rect.x() + bounding_rect.rect().width()
        min_y = bounding_rect.y()
        max_y = bounding_rect.y() + bounding_rect.rect().height()

        # Clamp position between min and max values
        x_position = min(max(relative_mouse_pos_x, min_x), max_x)
        y_position = min(max(relative_mouse_pos_y, min_y), max_y)

        # Do special check for if mouse is inside and dont move if so
        if (
            x_position < max_x
            and x_position > min_x
            and y_position < max_y
            and y_position > min_y
        ):
            return

        self.setX(x_position)
        self.setY(y_position)

        # Change orientation depending on what side the port is on
        if x_position == max_x:
            self.setOrientation(Orientation.RIGHT)
        elif x_position == min_x:
            self.setOrientation(Orientation.LEFT)
        elif y_position == max_y:
            self.setOrientation(Orientation.DOWN)
        elif y_position == min_y:
            self.setOrientation(Orientation.UP)

        super().mouseMoveEvent(event)
        self.moved.emit()

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        if event.button() == Qt.MouseButton.LeftButton:
            self.left_click = False

    def contextMenuEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        menu = QMenu()
        hide_signal_action = QAction("Toggle signal visibility", menu)
        hide_signal_action.triggered.connect(self.toggled)
        hide_signal_action.setEnabled(not self.isLocked)

        hide_port_action = QAction("Toggle port visibility", menu)
        hide_port_action.triggered.connect(self.toggleVisibility)
        hide_port_action.setEnabled(not self.isLocked)

        menu.addAction(hide_signal_action)
        menu.addAction(hide_port_action)
        menu.exec_(event.screenPos())
        event.accept()

    # TODO: Need comments here of what does what
    def toggleVisibility(self) -> None:
        self.setVisible(not self.isVisible())
        self.toggled.emit()

    def getDisplayName(self) -> str:
        return self.name

    def getSignalName(self) -> str:
        return self.signal.get_name()

    def getID(self) -> int:
        return self.id

    def setNameVisibility(self, is_visible):
        self.name_label.setVisible(is_visible)

    def setLocked(self, locked: bool) -> None:
        self.isLocked = locked

    def getOrientation(self) -> Orientation:
        return self.orientation
