import itertools
import math

from qtpy import QtCore
from qtpy.QtCore import QPointF, Qt
from qtpy.QtGui import QCursor
from qtpy.QtWidgets import QGraphicsItem, QGraphicsLineItem, QGraphicsSceneMouseEvent

from simudator.gui.color_scheme import ColorScheme
from simudator.gui.port_graphics_item import PortGraphicsItem

SHORTEST_LINE_LEN = 400
LONGEST_CLICK_DIST = 20
MERGE_THRESHOLD = 25


class SignalGraphicsItem(QGraphicsItem):
    """
    Class used to draw and update signals between modules, represented
    as orthogonal line segments that zig-zag from one module to another.
    The constituent line segments of the signal can both be moved orthogonally
    to the axis of the segments as well as be split into smaller segments.
    """

    id_iter = itertools.count()

    def __init__(
        self,
        start_port: PortGraphicsItem,
        end_port: PortGraphicsItem,
        parent=None,
        display_name: str = None,
    ):
        super().__init__(parent)
        self.start_port = start_port
        self.end_port = end_port
        self.id = next(self.id_iter)
        if display_name is None:
            self.name = start_port.getDisplayName() + "" + end_port.getDisplayName()
        else:
            self.name = display_name

        # Variables used for representing and drawing the signal
        self.points: list[QPointF] = []
        self.lines = []

        # Variables and flags for handling mouse events
        self.left_click = False
        self.last_line_start_i = None
        self.last_line_end_i = None
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsMovable)

        # Initialise the visual representation of the signal as a zig-zag
        # between the ports of the signal
        #           |---- port
        #           |
        # port -----|
        start_point = self.start_port.getPoint()
        end_point = self.end_port.getPoint()
        self.points.append(start_point)
        half_x = (start_point.x() + end_point.x()) / 2
        self.points.append(QPointF(half_x, start_point.y()))
        self.points.append(QPointF(half_x, end_point.y()))
        self.points.append(end_point)
        self.drawSignal()

        # Used to lock layout
        self.is_locked = False

    def reset(self) -> None:
        """
        Reset the visual representation of the signal to what it looked
        like when initialised.
        """
        self.points: list[QPointF] = []

        # Reset variables and flags for handling mouse events
        self.left_click = False
        self.last_line_start_i = None
        self.last_line_end_i = None

        # Initialise the visual representation of the signal as a zig-zag
        # between the ports of the signal
        #           |---- port
        #           |
        # port -----|

        start_point = self.start_port.getPoint()
        end_point = self.end_port.getPoint()
        self.points.append(start_point)
        half_x = (start_point.x() + end_point.x()) / 2
        self.points.append(QPointF(half_x, start_point.y()))
        self.points.append(QPointF(half_x, end_point.y()))
        self.points.append(end_point)

        self.setVisible(True)
        self.drawSignal()

    def drawSignal(self) -> None:
        """
        Draws the signal as a collection of orthogonal line segments which
        can be moved and split into smaller segments. Each segment is joined
        to another segment at its end points, except for the two end segments
        which are connected to module ports.
        """
        # Remove old lines from the GUI
        while self.lines:
            item = self.lines.pop()
            item.setParentItem(None)

        # Draw lines from the first to the second point, from the second
        # to the third point and so on
        prev_point = self.points[0]
        for point in self.points[1:]:
            # Coordinates
            # Bounded box for clicking
            x1 = prev_point.x()
            y1 = prev_point.y()
            x2 = point.x()
            y2 = point.y()
            line = QGraphicsLineItem(x1, y1, x2, y2, self)
            line.setPen(ColorScheme.Signal)
            self.lines.append(line)
            prev_point = point

    def splitLine(self, start_point_index: int, end_point_index: int) -> None:
        """
        Splits an existing line of two segments into a new line of four line
        segments by adding three new points and removing the old corer.

            |   ->    __|
        ____|   -> __|

        """

        # Last line is not splittable, split next last instead
        if end_point_index == len(self.points) - 1:
            start_point_index -= 1
            end_point_index -= 1

        # Get the points needed to do the splits and calculate the new points
        start_point = self.points[start_point_index]
        end_point = self.points[end_point_index]
        half_point = (start_point + end_point) / 2
        next_half_point = (end_point + self.points[end_point_index + 1]) / 2

        # Calculate the bot_corner point, it depends on the lines orientation
        if start_point.x() == end_point.x():
            bot_corner = QPointF(next_half_point.x(), half_point.y())
        else:
            bot_corner = QPointF(half_point.x(), next_half_point.y())

        # Remove the old corner and insteart new points in the correct order
        self.points.pop(end_point_index)
        self.points.insert(end_point_index, next_half_point)
        self.points.insert(end_point_index, bot_corner)
        self.points.insert(end_point_index, half_point)

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        """
        Handles pressing a line segment of the signal. The line segment
        that is closest to the point pressed will be chosen as the pressed
        segment. If the point pressed is too far away from the closest segment,
        no line segment will be chosen as pressed and the mouse press event
        is ignored.
        """
        # Calculate and save which line segment was pressed (choose the one
        # closest to the point pressed)
        press_point = event.scenePos()
        line_start_i, line_end_i = self.getClosestPoints(press_point)
        line_start = self.points[line_start_i]
        line_end = self.points[line_end_i]

        # Save which line segment was pressed
        self.last_line_start_i = min(line_start_i, line_end_i)
        self.last_line_end_i = max(line_start_i, line_end_i)

        # Save a bool representing whether we left clicked or not
        self.left_click = event.button() == Qt.MouseButton.LeftButton

        # Do not accept clicks too far away from the closest line segment
        dist_to_line = self.getLineDistance(line_start, line_end, press_point)
        if dist_to_line > LONGEST_CLICK_DIST:
            event.ignore()

        # Save whether the pressed line segment is vertical or horisontal
        # for convenience
        if line_start.x() == line_end.x():
            self.last_line_vertical = True
        elif line_start.y() == line_end.y():
            self.last_line_vertical = False

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        """
        Calls the functions 'handleLeftClick' or 'handleRightClick' if
        the user left or right clicked.
        """
        if event.button() == Qt.MouseButton.LeftButton:
            self.handleLeftClick()

        elif event.button() == Qt.MouseButton.RightButton:
            self.handleRightClick()

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        """
        Handles moving a pressed line segment with left click. The segment is
        moved horisontally of the segment itself is vertical and vice versa.
        The two line sgments that make up the ends of the signal are not moved
        if pressed.
        """
        if not self.left_click or self.is_locked:
            return

        # Change the cursor to become an orthogonal double pointed arrow
        if self.last_line_vertical:
            self.setCursor(QCursor(QtCore.Qt.SizeHorCursor))
        else:
            self.setCursor(QCursor(QtCore.Qt.SizeVerCursor))

        # Retrieve which line segment to move and where to move it
        current_mouse_point = event.scenePos()
        line_start = self.points[self.last_line_start_i]
        line_end = self.points[self.last_line_end_i]

        # Do not let the user move the start and end line segment connected
        # to the modules
        if line_start == self.points[0] or line_start == self.points[-1]:
            return
        if line_end == self.points[0] or line_end == self.points[-1]:
            return

        # If the line segment is vertical, move it horisontally
        if self.last_line_vertical:
            line_start.setX(current_mouse_point.x())
            line_end.setX(current_mouse_point.x())

        # If the line segment is horisontal, move it vertically
        if not self.last_line_vertical:
            line_start.setY(current_mouse_point.y())
            line_end.setY(current_mouse_point.y())

        self.drawSignal()

    def handleLeftClick(self) -> None:
        self.align(
            self.last_line_start_i, self.last_line_end_i, self.last_line_vertical
        )
        self.drawSignal()

        # Reset cursor
        self.setCursor(QCursor(QtCore.Qt.ArrowCursor))

    def align_search(self, start_index, ascending=False) -> int:
        """
        Recursivly searches deeper into the line structure to get the smallest
        and largest index of which all points in between should be removed to
        merge small line segments.
        """
        if ascending:
            if start_index + 2 >= len(self.points) - 1:
                return start_index

            distance_sq = self.distance_sq(
                self.points[start_index], self.points[start_index + 1]
            )

            if distance_sq < MERGE_THRESHOLD:
                return self.align_search(start_index + 2, ascending)
            return start_index

        else:
            if start_index - 2 <= 0:
                return start_index

            distance_sq = self.distance_sq(
                self.points[start_index], self.points[start_index - 1]
            )

            if distance_sq < MERGE_THRESHOLD:
                return self.align_search(start_index - 2, ascending)
            return start_index

    def align(self, line_start, line_end, vertical=False):
        """
        Merges smaller line segments that extend along the same axis into one
        larger line segment if the end points of the segments are close enough.
        """
        new_line_start = self.align_search(line_start, ascending=False)
        new_line_end = self.align_search(line_end, ascending=True)
        if vertical:
            self.points[new_line_end].setX(self.points[new_line_start].x())
        else:
            self.points[new_line_end].setY(self.points[new_line_start].y())
        self.points = self.points[: new_line_start + 1] + self.points[new_line_end:]

    def handleRightClick(self) -> None:
        """
        Splits an existing line into smaller ones using 'splitLine' if the
        line is long enough to be split.
        """
        line_start = self.points[self.last_line_start_i]
        line_end = self.points[self.last_line_end_i]

        # Do not let the user split line segments into too short segments
        if self.distance_sq(line_start, line_end) < SHORTEST_LINE_LEN:
            return

        # Do nothing if layoutlock is on
        if self.is_locked:
            return

        self.splitLine(self.last_line_start_i, self.last_line_end_i)
        self.drawSignal()

    def boundingRect(self):
        return self.childrenBoundingRect()

    def paint(self, *args):
        # This avoids errors, it is a pure virtual function that
        # needs an implementation
        pass

    def distance_sq(self, point1, point2):
        dx = point1.x() - point2.x()
        dy = point1.y() - point2.y()
        return dx**2 + dy**2

    def getClosestPoints(self, ref_point) -> tuple[int | None, int | None]:
        """
        Returns a tuple of the indexes to the points that make up the line
        segment that is closest to a reference point.
        """
        prev_point = self.points[0]
        shortest_distance = math.inf
        closest_1_i = None
        closest_2_i = None
        for index, point in enumerate(self.points[1:]):
            distance = self.getLineDistance(prev_point, point, ref_point)
            if distance < shortest_distance:  # Found a closer line
                shortest_distance = distance
                closest_1_i = index
                closest_2_i = index + 1

            prev_point = point

        return closest_1_i, closest_2_i

    def getLineDistance(self, line_start, line_end, ref_point) -> float:
        """
        Returns the distance from a reference point to the closest point on
        a given orthogonal line segment. If the closest point on the line
        segment lies outside the segment, the distance is considered
        infinite.
        """
        distance = math.inf
        # Vertical line segment
        if line_start.x() == line_end.x():
            # Make sure the closest point is not outside the line
            upper_y = max(line_start.y(), line_end.y())
            lower_y = min(line_start.y(), line_end.y())
            if lower_y <= ref_point.y() <= upper_y:
                distance = abs(ref_point.x() - line_end.x())

            else:
                distance = math.inf

        # Horisontal line segment
        elif line_start.y() == line_end.y():
            # Make sure the closest point is not outside the line
            leftmost_x = min(line_start.x(), line_end.x())
            rightmost_x = max(line_start.x(), line_end.x())
            if leftmost_x <= ref_point.x() <= rightmost_x:
                distance = abs(ref_point.y() - line_end.y())

            else:
                distance = math.inf

        return distance

    def move(self) -> None:
        """
        Updates the position of start or end line segments if the user has
        moved any of the modules that the signal is connected to.
        """
        # Find out whether the start or the end of the signal moved
        old_start = self.points[0]
        old_end = self.points[-1]
        new_start = self.start_port.getPoint()
        new_end = self.end_port.getPoint()

        # Index the points that should be moved
        if old_start != new_start:
            end_point = old_start
            neighbour_point = self.points[1]
            new_end_point = new_start
        elif old_end != new_end:
            end_point = old_end
            neighbour_point = self.points[-2]
            new_end_point = new_end
        else:
            return  # Nothing moved

        # Move the points correctly knowing the end points is horizontal
        if end_point.y() == neighbour_point.y():
            end_point.setX(new_end_point.x())
            end_point.setY(new_end_point.y())
            neighbour_point.setY(new_end_point.y())

        # Move the points correctly knowing the end points is vertical
        if end_point.x() == neighbour_point.x():
            end_point.setY(new_end_point.y())
            end_point.setX(new_end_point.x())
            neighbour_point.setX(new_end_point.x())

        self.drawSignal()

    def toggleVisibility(self) -> None:
        """
        Toggles the visibility of the signal.
        """
        self.setVisible(not self.isVisible())

    def getPoints(self) -> list[QPointF]:
        return self.points

    def getDisplayName(self) -> str:
        return self.name

    def getID(self) -> int:
        return self.id

    def setPoints(self, points: list[QPointF]) -> None:
        self.points = points

        # Redraw the signal
        self.drawSignal()
