from typing import Optional

from qtpy.QtCore import Qt
from qtpy.QtCore import Signal as pyqtSignal
from qtpy.QtGui import QValidator
from qtpy.QtWidgets import (
    QComboBox,
    QDialog,
    QHBoxLayout,
    QLineEdit,
    QPushButton,
    QWidget,
)

from simudator.core import Module


class HexValidator(QValidator):
    """
    A QValidator for constricting user input to a hex number that
    lies within a specified (exclusive) range.
    """

    def __init__(self, bottom: int, top: int) -> None:
        super().__init__(None)
        self.bottom = bottom
        self.top = top

    def validate(self, a0: str, a1: int) -> tuple['QValidator.State', str, int]:
        # A input string representing a hex number must begin with '0x'
        # The user must also be able to type '0x' if the input was empty before
        if a0[0:2] not in "0x":
            return (QValidator.Invalid, a0, a1)
        if not a0[2:]:
            return (QValidator.Intermediate, a0, a1)

        # The input string is converted to an int to check if it is a hex
        # number and to use later to check that it is within bounds
        try:
            parsed_int = int(a0, 16)
        except ValueError:
            return (QValidator.Invalid, a0, a1)

        # The input string converted to a hex number must be withing the
        # specified range of this validor to be accepted
        else:
            if self.bottom <= parsed_int < self.top:
                return (QValidator.Acceptable, a0, a1)
            return (QValidator.Invalid, a0, a1)


class MiaMemoryContentDialog(QDialog):
    accepted = pyqtSignal(str, str, str, name="accepted")

    def __init__(
        self,
        memory: Module,
        parent: Optional['QWidget'] = None,
        flags: Qt.WindowFlags | Qt.WindowType = Qt.WindowFlags(),
    ) -> None:
        super().__init__(parent, flags)

        self.memory = memory
        self.memory_name = memory.get_state()['name']

        adresses = [hex(i) for i in range(len(memory.get_state()['memory']))]
        adressValidator = HexValidator(0, len(adresses))

        self.adressSelectWidget = QComboBox()
        self.adressSelectWidget.addItems(adresses)
        self.adressSelectWidget.activated.connect(self.updateValue)
        self.adressSelectWidget.setEditable(True)
        self.adressSelectWidget.setValidator(adressValidator)

        self.valuesWidget = QLineEdit()

        okButton = QPushButton('OK')
        okButton.clicked.connect(self.signalAccepted)
        cancelButton = QPushButton('Cancel')
        cancelButton.clicked.connect(self.close)

        hboxLayout = QHBoxLayout()
        hboxLayout.addWidget(self.adressSelectWidget, 0)
        hboxLayout.addWidget(self.valuesWidget, 1)
        hboxLayout.addWidget(okButton)
        hboxLayout.addWidget(cancelButton)
        self.setLayout(hboxLayout)

        self.show()
        self.updateValue()

    def updateValue(self) -> None:
        selectedAdress = int(self.adressSelectWidget.currentText(), 16)
        memory_content = self.memory.get_state()['memory']
        value = memory_content[selectedAdress]
        self.valuesWidget.setText(hex(value))

    def signalAccepted(self) -> None:
        adress = self.adressSelectWidget.currentText()
        enteredValues = self.valuesWidget.text()
        self.accepted.emit(self.memory_name, adress, enteredValues)
        self.close()
