from simudator.core.module import Module
from simudator.core.signal import Signal


class Bus(Module):
    """
    A module for a bus which has an arbitrary amount of input signals and
    output signals. When a value other than None is sent onto the bus from
    an input signal, the value is sent onto all output signals. Only one
    input signal can send a value to the bus at a time.
    """

    __slots__ = ()

    def __init__(
        self, inputs: list[Signal] = [], outputs: list[Signal] = [], name: str = "Bus"
    ) -> None:

        signals = {}
        for i, s in enumerate(inputs):
            signals[f"in_input_{i}"] = s
        for i, s in enumerate(outputs):
            signals[f"out_output_{i}"] = s
        super().__init__(signals, name)

    def update_logic(self) -> None:
        """
        Transfer the input value onto all output signals. The input value
        is the value of whichever input signal that has a value other than
        None. If more than two input signals have a value other than None,
        nothing will be transferred onto the output signals.
        """
        value = None
        for input_signal in self.get_input_signals():
            if input_signal.get_value() is not None:
                if value is not None:
                    # Two input signals are writing to the bus at the
                    # same time
                    value = None
                    break
                value = input_signal.get_value()

        if value is not None:
            for output_signal in self.get_output_signals():
                output_signal.update_value(value)

    def get_state(self) -> dict:
        return super().get_state()

    def set_state(self, state: dict) -> None:
        super().set_state(state)

    def print_module(self) -> None:
        pass

    def reset(self) -> None:
        pass

    def get_name(self) -> str:
        return self.name

    def update_register(self) -> None:
        """
        The bus has no register.
        """
        pass

    def output_register(self) -> None:
        """
        The bus has no register.
        """
        pass

    def save_state_to_file(self, file_path: str) -> bool:
        """
        The ali is state-less.
        """
        return True
