from typing import Any

from simudator.core.module import Module
from simudator.core.signal import Signal
from simudator.processor.mia.modules.mia_bus_connect import MiaBusConnector


class IR(Module, MiaBusConnector):
    """
    A MIA specific register containing the current instruction to
    be performed by the cpu. The instructions are dividied into
    several fields.
    """

    # Python does not allow multiple inherintence if more than one of the
    # parent classes uses __slots__. Thus we also includes the __slots__
    # from MiaBusConnector.
    __slots__ = ("op", "grx", "m", "a", "bus_id", "instruction", "bus_control_s")

    def __init__(
        self,
        to_bus: Signal,
        from_bus: Signal,
        op_s: Signal,
        grx_s: Signal,
        m_s: Signal,
        bus_control: Signal,
        default_value=0,
        bus_id=0,
        name: str = "IR",
    ) -> None:

        signals = {
            "in_input": from_bus,
            "out_output": to_bus,
            "out_op": op_s,
            "out_grx": grx_s,
            "out_m": m_s,
        }
        Module.__init__(self, signals, name)
        MiaBusConnector.__init__(self, bus_control, bus_id)

        # Internal values
        self.op = 0
        self.grx = 0
        self.m = 0
        self.a = 0
        self.bus_id = bus_id
        self.instruction = default_value

    def update_register(self) -> None:
        """
        Updates the registers internal values by reading the value
        from the bus and bitshift each part of the instruction.
        """
        if self.read_from_bus():
            self.instruction = self.signals["in_input"].get_value()
            self.op = self.instruction >> 12
            self.grx = (self.instruction >> 10) & 0b11
            self.m = (self.instruction >> 8) & 0b11
            self.a = self.instruction & (2**8 - 1)

    def output_register(self) -> None:
        """
        Always output the internal values to the signals op_s, grx_s
        and m_s, and output its whole instruction to the bus when
        asked to.
        """
        self.signals["out_op"].update_value(self.op)
        self.signals["out_grx"].update_value(self.grx)
        self.signals["out_m"].update_value(self.m)

        if self.write_to_bus():
            self.signals["out_output"].update_value(self.instruction)
        else:
            self.signals["out_output"].update_value(None)

    def update_logic(self) -> None:
        self.output_register()

    def get_state(self) -> dict[str, Any]:
        state = {
            "name": self.name,
            "value": self.instruction,
            "op": self.op,
            "grx": self.grx,
            "m": self.m,
            "a": self.a,
        }
        return state

    def get_gui_state(self) -> dict:
        state = super().get_gui_state()
        state["value"] = self.instruction
        return state

    def set_state(self, state: dict) -> None:
        super().set_state(state)
        self.instruction = state["value"]
        if "op" in state:
            self.op = state["op"]
        if "grx" in state:
            self.grx = state["grx"]
        if "m" in state:
            self.m = state["m"]
        if "a" in state:
            self.a = state["a"]

    def reset(self) -> None:
        """
        Resets the instruction register to 0.
        """
        self.instruction = 0
        self.op = 0
        self.grx = 0
        self.m = 0
        self.a = 0

    def print_module(self) -> None:
        print(
            "",
            self.name,
            "\n -----",
            "\n value: ",
            hex(self.instruction),
            "\n op: ",
            bin(self.op),
            "\n grx: ",
            bin(self.grx),
            "\n m: ",
            bin(self.m),
            "\n a: ",
            bin(self.a),
        )

    def save_state_to_file(self, file_path: str) -> None:
        """
        Tries to save the modules state to a given file.
        """
        file = open(file_path, "a")
        file.write(self.name + ":\n")
        file.write("Instruction: " + hex(self.instruction)[2:] + "\n\n")
        file.close()

    def load_from_str(self, state_string):
        string_pair = state_string.split(": ")
        # TODO: Maybe check if it starts with instruction: ?
        self.instruction = int(string_pair[1], 16)
