from __future__ import annotations

from typing import Any

from simudator.core.module import Module
from simudator.core.signal import Signal
from simudator.processor.mia.modules.mia_bus_connect import MiaBusConnector


class GRX(Module, MiaBusConnector):
    """
    A class representing the four registers and the mux/demux
    used in the mia architecture. The S signal determines if the
    registers should be indexed by the GRx bits or the M bits.
    """

    # Python does not allow multiple inherintence if more than one of the 
    # parent classes uses __slots__. Thus we also includes the __slots__
    # from MiaBusConnector.
    __slots__ = (
        "grx_control",
        "m_control",
        "s_control",
        "bus_id",
        "registers",
        "bus_control_s",
    )

    def __init__(
        self,
        to_bus: Signal,
        from_bus: Signal,
        bus_control: Signal,
        grx_control: Signal,
        m_control: Signal,
        s_control: Signal,
        bus_id: int,
        name="GRx",
        registers=[0 for _ in range(4)],
    ) -> None:
        # Set connection to/from bus and bus_id
        MiaBusConnector.__init__(self, bus_control, bus_id)

        # Other signals
        signals = {
            "in_input": from_bus,
            "in_control_grx": grx_control,
            "in_control_s": s_control,
            "in_control_m": m_control,
            "out_content": to_bus,
        }
        # Init module superclass
        Module.__init__(self, signals, name)

        # set the registers
        self.registers = registers

    def update_register(self):
        """
        If the module should read from the bus, update the correct
        register by reading the s bit and index the registers by
        the GRx bits if s == 0 or M bits if s == 1.
        """
        if self.read_from_bus():
            s_control_value = self.signals["in_control_s"].get_value()

            grx_index = 0
            if s_control_value == 0b0:
                grx_index = self.signals["in_control_grx"].get_value()
            elif s_control_value == 0b1:
                grx_index = self.signals["in_control_m"].get_value()
            else:
                raise ValueError("The value of the s signal" "must be 0b0 or 0b1 !!!")

            self.registers[grx_index] = self.signals["in_input"].get_value()

    def output_register(self):
        """
        If the module should write to the bus, update the bus
        by reading the s bit and index the registers by
        the GRx bits if s == 0 or M bits if s == 1.
        """
        if self.write_to_bus():
            s_control_value = self.signals["in_control_s"].get_value()

            grx_index = 0
            if s_control_value == 0b0:
                grx_index = self.signals["in_control_grx"].get_value()
            elif s_control_value == 0b1:
                grx_index = self.signals["in_control_m"].get_value()
            else:
                raise ValueError("The value of the s signal" "must be 0b0 or 0b1 !!!")

            self.signals["out_content"].update_value(self.registers[grx_index])
        else:
            self.signals["out_content"].update_value(None)

    def update_logic(self) -> None:
        self.output_register()

    def get_state(self) -> dict[str, Any]:
        """
        Returns a dict of the grx state.
        These states are changable via set_states.
        """
        state = {
            "name": self.name,
            "registers": self.registers[:],
        }
        return state

    def get_gui_state(self) -> dict:
        state = {
            "name": self.name,
            "registers": self.registers[:],
        }
        return state

    def set_state(self, state: dict[str, Any]):
        """
        Sets the grx state to one given in dict.
        """
        self.name = state["name"]
        self.registers = state["registers"]

    def reset(self) -> None:
        """
        Resets the GRx registers to 0.
        """
        for i in range(len(self.registers)):
            self.registers[i] = 0

    def save_state_to_file(self, file_path: str) -> None:
        """
        Tries to save the modules state to a given file.
        Returns true on success and false if something went wrong.
        """
        file = open(file_path, "a")
        file.write(self.name + ":\n")
        for index, value in enumerate(self.registers):
            file.write(hex(index)[2:] + ": ")
            file.write(hex(value)[2:])
            file.write("\n")

        file.write("\n")
        file.close()

    def load_from_str(self, state_string) -> None:
        """
        Loads the module from a string, used when loading state from
        files.
        """
        self.reset()
        lines = state_string.split("\n")
        for line in lines:
            if line:  # last line is empty from split with \n
                line_data = line.split(": ")
                adress = int(line_data[0], 16)
                value = int(line_data[1], 16)
                self.registers[adress] = value

    def print_module(self):
        print(
            "",
            self.name,
            "\n -----",
            "\n bus_id: ",
            self.bus_id,
            "\n GR0: ",
            hex(self.registers[0]),
            "\n GR1: ",
            hex(self.registers[1]),
            "\n GR2: ",
            hex(self.registers[2]),
            "\n GR3: ",
            hex(self.registers[3]),
        )
