from simudator.core.modules.memory import Memory
from simudator.core.signal import Signal
from simudator.processor.mia.modules.mia_bus_connect import MiaBusConnector


class MiaMemory(MiaBusConnector, Memory):
    """
    A MIA specific memory, functionality from the general memory
    and the MiaBusConnector is inherited. The module can be of
    arbitrary size and supports opperations that read and write
    to/from the mia bus.
    """

    # Python does not allow multiple inherintence if more than one of the
    # parent classes uses __slots__. Thus we also includes the __slots__
    # from MiaBusConnector.
    __slots__ = ("label_adress_mapping", "bus_id", "bus_control_s")

    def __init__(
        self,
        input_signal: Signal,
        output_signal: Signal,
        address_signal: Signal,
        bus_control_s: Signal,
        size: int = 1,
        bus_id: int = 0,
        name: str = "PM",
    ) -> None:
        MiaBusConnector.__init__(self, bus_control_s, bus_id)
        Memory.__init__(
            self,
            input_signal,
            output_signal,
            Signal(None),
            address_signal,
            size,
            name=name,
        )
        self.bus_control_s.add_destination(self)

        self.label_adress_mapping = {}

    def update_register(self) -> None:
        """
        Updates the value of the current address.
        """
        if self.read_from_bus():
            self._memory[self._current_address] = self.signals["in_input"].get_value()

    def output_register(self) -> None:
        """
        Outputs the value of the current address.
        """
        if self.write_to_bus():
            self.signals["out_content"].update_value(
                self._memory[self._current_address]
            )
        else:
            self.signals["out_content"].update_value(None)

    def update_logic(self) -> None:
        """
        Outputs the value of the current address unless current address
        is None
        """
        address_value = self.signals["in_address"].get_value()
        if address_value is not None:
            self._current_address = address_value
            if self.write_to_bus():
                self.signals["out_content"].update_value(
                    self._memory[self._current_address]
                )
            else:
                self.signals["out_content"].update_value(None)

    def reset(self) -> None:
        size = len(self._memory)
        self._memory = [0 for _ in range(size)]

    def load_from_str(self, state_string) -> None:
        """
        Loads the module from a string, used when loading state from
        files.
        """
        self.reset()
        lines = state_string.split("\n")
        for line in lines:

            if line:  # last line is empty from split with \n
                line_data = line.split()
                value = int(line_data[1], 16)

                # Last character of the adress is a semicolon
                adress = int(line_data[0][:-1], 16)

                self._memory[adress] = value

                # There is an asm instruction label
                if len(line_data) == 3:
                    instr = line_data[2]
                    self.label_adress_mapping[adress] = instr

    def get_label(self, adress: int) -> str:
        """Return the label at the given memory address if it exists, else an empty string"""

        if adress in self.label_adress_mapping.keys():
            return self.label_adress_mapping[adress]

        return ""
