from typing import Any

from simudator.core.module import Module
from simudator.core.signal import Signal


class MicroMemory(Module):
    """An implementation of the micro memory from the MIA lab. The micro memory
    has control signals for the ALU, bus transfer, the general registers GRx,
    the program counter PC, the loop counter LC and the micro program counter
    uPC. It also has signals for sending data to the bus, the LC and the uPC as
    well as signals for reading the flags z, n, c, o and l.

    The bus transfer control signal is intended to be connected to all modules
    which will read or write to the bus. The signal value controls which two
    modules that respectively write to and read from the bus during a clock
    cycle.
    """

    __slots__ = (
        "curr_instr",
        "z_flag_val",
        "n_flag_val",
        "c_flag_val",
        "o_flag_val",
        "l_flag_val",
        "memory",
        "halt",
        "adress_padding",
        "value_padding",
    )

    def __init__(
        self,
        alu: Signal,
        bus_control: Signal,
        bus_constant: Signal,
        grx_control: Signal,
        pc_control: Signal,
        lc_control: Signal,
        lc_uadr: Signal,
        upc: Signal,
        upc_control: Signal,
        upc_uadr: Signal,
        z_flag: Signal,
        n_flag: Signal,
        c_flag: Signal,
        o_flag: Signal,
        l_flag: Signal,
        name: str = "uM",
        adress_padding: int = 2,
        value_padding: int = 7,
    ) -> None:
        """
        Parameters
        ----------
        alu : Signal
            Control signal specifying the ALU operand.
        bus_control : Signal
            Control signal for which modules should write to and read
            from the bus.
        bus_constant : Signal
            Signal for sending uADR instruction field to bus.
        grx_control : Signal
            Control signal for GRx.
        pc_control : Signal
            Control signal for program counter PC.
        lc_control : Signal
            Control signal for loop counter LC.
        lc_uadr : Signal
            Signal for sending uADR instruction field to LC.
        upc : Signal
            Signal carrying the value of the micro program counter uPC.
        upc_control : Signal
            Control signal for uPC.
        upc_uadr : Signal
            Signal for sending uADR instruction field to uPC.
        z_flag : Signal
            Signal carrying the value of the z-flag.
        n_flag : Signal
            Signal carrying the value of the n-flag.
        c_flag : Signal
            Signal carrying the value of the c-flag.
        o_flag : Signal
            Signal carrying the value of the o-flag.
        l_flag : Signal
            Signal carrying the value of the l-flag.
        name : str
            Optional name for the module. Defaults to 'uM'.

        Returns
        -------
        unknown
            None
        """

        signals = {
            "in_upc": upc,
            "in_flag_z": z_flag,
            "in_flag_n": n_flag,
            "in_flag_c": c_flag,
            "in_flag_o": o_flag,
            "in_flag_l": l_flag,
            "out_alu": alu,
            "out_grx": grx_control,
            "out_pc": pc_control,
            "out_bus": bus_constant,
            "out_upc": upc_uadr,
            "out_lc": lc_uadr,
            "out_control_bus": bus_control,
            "out_control_lc": lc_control,
            "out_control_upc": upc_control,
        }

        super().__init__(signals, name)

        # Internal variables
        self.curr_instr = 0
        self.z_flag_val = 0
        self.n_flag_val = 0
        self.c_flag_val = 0
        self.o_flag_val = 0
        self.l_flag_val = 0
        self.memory = [0 for i in range(128)]
        self.halt = False  # Used for signalling a HALT

        # paddings to ensure saving to file has the correct format
        self.adress_padding = adress_padding
        self.value_padding = value_padding

    def update_logic(self) -> None:
        """
        Update the micro controller with a new instruction index.

        The
        micro controller sets control signals to other modules in accordance
        with the micro instruction pointed to.
        """
        instruction = self.signals["in_upc"].get_value()
        if instruction is None:
            return

        self.curr_instr = instruction
        self.z_flag_val = self.signals["in_flag_z"].get_value()
        self.n_flag_val = self.signals["in_flag_n"].get_value()
        self.c_flag_val = self.signals["in_flag_c"].get_value()
        self.o_flag_val = self.signals["in_flag_o"].get_value()
        self.l_flag_val = self.signals["in_flag_l"].get_value()

        # Extract the different fields of the micro instruction
        instr = self.memory[self.curr_instr]
        alu_field = (instr >> 21) & 0b1111
        tb_field = (instr >> 18) & 0b111
        fb_field = (instr >> 15) & 0b111
        s_field = (instr >> 14) & 0b1
        p_field = (instr >> 13) & 0b1
        lc_field = (instr >> 11) & 0b11
        seq_field = (instr >> 7) & 0b1111
        uadr_field = instr & 0b1111111

        # Special case where only the ALU field and the TB field of
        # the microinstruction should be used
        if tb_field == 0b111:
            # Send bit 9-24 of the instruction onto the bus to be used
            # by the ALU
            # Increase uPC by 1
            instr_constant = instr & 0b1111111111111111
            self.signals["out_alu"].update_value(alu_field)
            self.signals["out_control_bus"].update_value((0b000, 0b000))
            self.signals["out_bus"].update_value(instr_constant)
            self.signals["out_control_upc"].update_value(0)

            # Make sure other control signals are set to 0 so that there
            # is no unexpected behaviour from leftover signal values
            self.signals["out_control_lc"].update_value(0)
            self.signals["out_pc"].update_value(0)
            self.signals["out_grx"].update_value(0)
            return

        if (0b0001 <= alu_field <= 0b0010) or (0b0100 <= alu_field <= 0b1000):
            # Special case for writing to the ALU from the bus, in which case
            # the FB field of the instruction is ignored (overwritten, even)
            # (The ALU does not have an actual bus adress in the MIA processor)
            fb_field = 0b1000

        # Handle LC according to the LC field
        match lc_field:
            case 0b00:
                self.signals["out_control_lc"].update_value(0b00)
            case 0b01:
                self.signals["out_control_lc"].update_value(0b01)
            case 0b10:
                self.signals["out_control_lc"].update_value(0b10)
                # fb_field = 0b1001 # Ignores actual FB field
            case 0b11:
                self.signals["out_control_lc"].update_value(0b11)
                self.signals["out_lc"].update_value(uadr_field & 0b01111111)

        # Update control signals that require no special handling
        self.signals["out_alu"].update_value(alu_field)
        self.signals["out_bus"].update_value(None)
        self.signals["out_control_bus"].update_value((tb_field, fb_field))
        self.signals["out_grx"].update_value(s_field)
        self.signals["out_pc"].update_value(p_field)

        # Handle uPC according to the SEQ field
        match seq_field:
            case 0b0000:
                self.signals["out_control_upc"].update_value(0b000)
            case 0b0001:
                self.signals["out_control_upc"].update_value(0b001)
            case 0b0010:
                self.signals["out_control_upc"].update_value(0b010)
            case 0b0011:
                self.signals["out_control_upc"].update_value(0b011)
            case 0b0100:
                self._conditional_jump(self.z_flag_val, 0, uadr_field)
            case 0b0101:
                self.signals["out_control_upc"].update_value(0b100)
                self.signals["out_upc"].update_value(uadr_field)
            case 0b0110:
                self.signals["out_control_upc"].update_value(0b110)
                self.signals["out_upc"].update_value(uadr_field)
            case 0b0111:
                self.signals["out_control_upc"].update_value(0b101)
            case 0b1000:
                self._conditional_jump(self.z_flag_val, 1, uadr_field)
            case 0b1001:
                self._conditional_jump(self.n_flag_val, 1, uadr_field)
            case 0b1010:
                self._conditional_jump(self.c_flag_val, 1, uadr_field)
            case 0b1011:
                self._conditional_jump(self.o_flag_val, 1, uadr_field)
            case 0b1100:
                self._conditional_jump(self.l_flag_val, 1, uadr_field)
            case 0b1101:
                self._conditional_jump(self.c_flag_val, 0, uadr_field)
            case 0b1110:
                self._conditional_jump(self.o_flag_val, 0, uadr_field)
            case 0b1111:
                # Halt is handled by update_register
                self.signals["out_control_upc"].update_value(0b011)

    def update_register(self) -> None:
        """
        Signal halt when the micro memory performs a halt instruction.

        Notes
        -----
        Although the micro memory is not a register, this override is needed to
        make sure a halt instruction is signalled on the actual clock cycle
        where it is "executed". All control signals for the halt will have been
        propagated at the end of the previous cycle. If the halt signalling
        were to be done in update_logic, the halt would erroneously be signaled
        one cycle to early.
        """
        seq_field = (self.memory[self.curr_instr] >> 7) & 0b1111
        if seq_field == 0b1111:
            self.halt = True
        else:
            self.halt = False

    def _conditional_jump(self, flag, cond_value, uadr):
        """Helper function for executing a conditional jump to the specified uadr
        if the specified flag has the specified value. If the condition is not
        true then uPC is increased by 1.
        """
        if flag == cond_value:
            self.signals["out_control_upc"].update_value(0b100)
            self.signals["out_upc"].update_value(uadr)
            return
        self.signals["out_control_upc"].update_value(0b000)

    def print_module(self) -> None:
        """Print the contents of the micro memory.

        Adresses which contain the
        value 0 are not printed.
        """
        print("", self.name, "\n -----\n")
        for adress in range(len(self.memory)):
            if self.memory[adress] != 0:
                print("", str(adress), "", str(self.memory[adress]), "\n")

    def get_state(self) -> dict:
        state = super().get_state()
        state["memory"] = self.memory[:]
        state["halt"] = self.halt
        state["curr_instr"] = self.curr_instr
        return state

    def get_gui_state(self) -> dict[str, Any]:
        state = super().get_state()
        state["memory"] = self.memory[:]
        return state

    def set_state(self, state: dict) -> None:
        super().set_state(state)
        self.memory = state["memory"]
        if "halt" in state:
            self.halt = state["halt"]

    def output_register(self):
        pass

    def get_uPC_signal(self):
        return self.signals["in_upc"]

    def reset(self) -> None:
        self.curr_instr = 0
        self.z_flag_val = 0
        self.n_flag_val = 0
        self.c_flag_val = 0
        self.o_flag_val = 0
        self.l_flag_val = 0
        self.memory = [0 for i in range(128)]
        self.halt = False  # Used for signalling a HALT

    def load_from_str(self, state_string) -> None:
        """Load the contents of the micro memory from a string consisting of a
        row for every adress. Each row should be formatted as 'ADR: VALUE'.
        Both the adress and the value should be written in hex notation.
        Adresses which are not specified in the string will receive the value
        0.

        Example string where all adresses except for 0x00, 0x01 and 0xb0
        receive the value 0 by default::

            '
            00: 0001
            01: 00a4
            b0: ffff
            '
        """
        for i in range(len(self.memory)):
            self.memory[i] = 0

        lines = state_string.split("\n")
        for line in lines:
            if line:  # last line is empty from split with \n
                line_data = line.split(": ")
                adress = int(line_data[0], 16)
                value = int(line_data[1], 16)
                self.memory[adress] = value

    def save_state_to_file(self, file_path: str) -> None:
        """Tries to save the modules state to a given file."""
        file = open(file_path, "a")
        file.write(self.name + ":\n")
        for index, value in enumerate(self.memory):
            # Write the adress in hex and add ': ' to the end
            file.write(hex(index)[2:].rjust(self.adress_padding, "0") + ": ")
            # Write the value in hex
            file.write(hex(value)[2:].rjust(self.value_padding, "0"))
            file.write("\n")

        file.write("\n")
        file.close()

    def get_largest_mem_adr(self) -> int:
        """Helper function for pretty_print that returns the length of
        the largest adress in the memory to print for a module.
        """
        return len(str(len(self.memory)))

    def get_longest_line_len(self, ignore_keys=[]) -> int:
        """Helper function for pretty_print that returns the length of
        the longest value in the memory to print for a module.
        """

        longest_memory_line = 0

        for value in self.memory:
            value_len = len(str(value))
            if value_len > longest_memory_line:
                longest_memory_line = value_len

        return longest_memory_line
